from typing import Any
from typing import Dict


async def update_distribution(
    hub,
    ctx,
    name: str,
    before: Dict[str, Any],
    resource_id: str,
    caller_reference: str,
    origins: Dict,
    default_cache_behaviour: Dict[str, Any],
    comment: str,
    logging: Dict[str, Any],
    enabled: str,
    viewer_certificate: str,
    aliases: Dict,
    price_class: str,
    default_root_object: str,
    origin_groups: Dict[str, Any],
    cache_behaviors: Dict[str, Any],
    custom_error_responses: Dict[str, Any],
    restrictions: Dict[str, Any],
    web_acl_id: str,
    http_version: str,
    is_ipv6_enabled: bool,
    if_match: str,
    timeout: Dict[str, Any],
):
    result = dict(comment=(), result=True, ret=None)
    # cloudfront distribution requires all params to be sent for update call even though one parameter change.
    # we will check if at least one attribute is changed and send the update request if any of the attribute changes.

    update_allowed_params = [
        "caller_reference",
        "origins",
        "default_cache_behaviour",
        "comment",
        "logging",
        "enabled",
        "viewer_certificate",
        "aliases",
        "price_class",
        "default_root_object",
        "origin_groups",
        "cache_behaviors",
        "custom_error_responses",
        "restrictions",
        "web_acl_id",
        "http_version",
        "is_ipv6_enabled",
    ]

    modified_params = {}
    for parameter in update_allowed_params:
        if (
            locals()[parameter] is not None
            and before.get(parameter) != locals()[parameter]
        ):
            modified_params[parameter] = locals()[parameter]

    if modified_params:
        result["ret"] = modified_params
        if not ctx.get("test", False):
            distribution_config = {
                "CallerReference": caller_reference,
                "Origins": origins,
                "DefaultCacheBehavior": default_cache_behaviour,
                "Comment": comment,
                "Logging": logging,
                "Enabled": enabled,
                "ViewerCertificate": viewer_certificate,
                "Aliases": aliases,
                "PriceClass": price_class,
                "DefaultRootObject": default_root_object,
                "OriginGroups": origin_groups,
                "CacheBehaviors": cache_behaviors,
                "CustomErrorResponses": custom_error_responses,
                "Restrictions": restrictions,
                "WebACLId": web_acl_id,
                "HttpVersion": http_version,
                "IsIPV6Enabled": is_ipv6_enabled,
            }
            # Even though some params are optional in create, most of them are mandatory (at least default values) in
            # update call. Refer
            # https://docs.aws.amazon.com/AmazonCloudFront/latest/DeveloperGuide/distribution-overview-required-fields.html
            # we will sanitize the distribution config and fill in missing fields with default values.

            distribution_config = hub.tool.aws.cloudfront.distribution_utils.sanitize_distribution_config_for_update(
                distribution_config
            )

            ret = await hub.exec.boto3.client.cloudfront.update_distribution(
                ctx,
                DistributionConfig=distribution_config,
                Id=resource_id,
                IfMatch=if_match,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

            waiter_ret = await hub.tool.aws.cloudfront.distribution.distribution_waiter(
                ctx, name, resource_id, timeout, "update"
            )
            if not waiter_ret["result"]:
                result["result"] = False
                result["comment"] = result["comment"] + waiter_ret["comment"]
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.cloudfront.distribution", name=name
            )
    return result


async def disable_distribution(
    hub,
    ctx,
    name: str,
    resource_id: str,
    before: Dict[str, Any],
    if_match: str,
    timeout: Dict[str, Any],
):
    return await hub.tool.aws.cloudfront.distribution.update_distribution(
        ctx=ctx,
        name=name,
        before=before,
        resource_id=resource_id,
        caller_reference=before.get("caller_reference"),
        origins=before.get("origins"),
        default_cache_behaviour=before.get("default_cache_behaviour"),
        comment=before.get("comment"),
        logging=before.get("logging"),
        enabled=False,
        viewer_certificate=before.get("viewer_certificate"),
        aliases=before.get("aliases"),
        price_class=before.get("price_class"),
        default_root_object=before.get("default_root_object"),
        origin_groups=before.get("origin_groups"),
        cache_behaviors=before.get("cache_behaviors"),
        custom_error_responses=before.get("custom_error_responses"),
        restrictions=before.get("restrictions"),
        web_acl_id=before.get("web_acl_id"),
        http_version=before.get("http_version"),
        is_ipv6_enabled=before.get("is_ipv6_enabled"),
        if_match=if_match,
        timeout=timeout,
    )


async def distribution_waiter(
    hub, ctx, name: str, resource_id: str, timeout: Dict, operation_type: str
):
    """

    Waiter to wait for the distribution to become active.

        Args:
           hub:
           ctx:
           name(Text): An Idem name of the resource.
           resource_id(Text): Id of the distribution to Identify the resource
           timeout(Dict, optional): Timeout configuration for creating or updating distribution.
            * create (Dict) -- Timeout configuration for creating distribution
                * delay(int, default=60) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=40) -- Customized timeout configuration containing delay and max attempts.
            * update (string) -- Timeout configuration for updating distribution
                * delay(int, default=60) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=40) -- Customized timeout configuration containing delay and max attempts.
           operation_type(Text): create or update operation

        Returns:
            Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret={})
    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=60,
        default_max_attempts=35,
        timeout_config=timeout.get(operation_type) if timeout else None,
    )
    hub.log.debug(f"Waiting on {operation_type} aws.cloudfront.distribution '{name}'")
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "cloudfront",
            "distribution_deployed",
            Id=resource_id,
            WaiterConfig=waiter_config,
        )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result



import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub,
    ctx,
    key_id: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """

    Args:
        hub:
        ctx:
        key_id: aws kms key id
        old_tags: dict of old tags
        new_tags: dict of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": "dict of tags after update"}

    """
    result = dict(comment=(), result=True, ret=None)

    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove and not ctx.get("test", False):
        tag_keys = list(tags_to_remove.keys())
        delete_ret = await hub.exec.boto3.client.kms.untag_resource(
            ctx, KeyId=key_id, TagKeys=tag_keys
        )
        if not delete_ret["result"]:
            result["comment"] = delete_ret["comment"]
            result["result"] = False
            return result
    if tags_to_add and not ctx.get("test", False):
        add_ret = await hub.exec.boto3.client.kms.tag_resource(
            ctx,
            KeyId=key_id,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list_tagkey(
                tags=tags_to_add
            ),
        )
        if not add_ret["result"]:
            result["comment"] = add_ret["comment"]
            result["result"] = False
            return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_id,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """Update tags of AWS CloudTrail resources.

    Args:
        resource_id: aws resource id
        old_tags: dict of old tags in the format of {tag-key: tag-value}
        new_tags: dict of new tags in the format of {tag-key: tag-value}. If this value is None, the function will do no operation on tags.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": dict of updated tags}

    """
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.cloudtrail.remove_tags(
                ctx,
                ResourceId=resource_id,
                TagsList=hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                    tags=tags_to_remove
                ),
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result

    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.cloudtrail.add_tags(
                ctx,
                ResourceId=resource_id,
                TagsList=hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                    tags=tags_to_add
                ),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result


async def update_trail(
    hub,
    ctx,
    before: Dict[str, Any],
    s3_bucket_name: str,
    s3_key_prefix: str,
    sns_topic_name: str,
    include_global_service_events: bool,
    is_multi_region_trail: bool,
    enable_logfile_validation: bool,
    cloud_watch_logs_loggroup_arn: str,
    cloud_watch_logs_role_arn: str,
    kms_key_id: str,
    is_organization_trail: bool,
):
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}

    resource_parameters = OrderedDict(
        {
            "S3BucketName": s3_bucket_name,
            "S3KeyPrefix": s3_key_prefix,
            "SnsTopicName": sns_topic_name,
            "IncludeGlobalServiceEvents": include_global_service_events,
            "IsMultiRegionTrail": is_multi_region_trail,
            "CloudWatchLogsLogGroupArn": cloud_watch_logs_loggroup_arn,
            "CloudWatchLogsRoleArn": cloud_watch_logs_role_arn,
            "KmsKeyId": kms_key_id,
            "IsOrganizationTrail": is_organization_trail,
        }
    )

    if "LogFileValidationEnabled" in before.keys():
        if (
            enable_logfile_validation is not None
        ) and enable_logfile_validation != before["LogFileValidationEnabled"]:
            update_payload["EnableLogFileValidation"] = enable_logfile_validation

    for key, value in resource_parameters.items():
        if key in before.keys():
            if (value is not None) and value != before[key]:
                update_payload[key] = resource_parameters[key]

    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.cloudtrail.update_trail(
                ctx, Name=before.get("Name"), **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        result = update_result(result, update_payload)

    return result


def update_result(
    result: Dict[str, Any], update_payload: Dict[str, Any]
) -> Dict[str, Any]:
    parameters = OrderedDict(
        {
            "S3BucketName": "s3_bucket_name",
            "S3KeyPrefix": "s3_key_prefix",
            "SnsTopicName": "sns_topic_name",
            "IncludeGlobalServiceEvents": "include_global_service_events",
            "IsMultiRegionTrail": "is_multi_region_trail",
            "EnableLogFileValidation": "enable_logfile_validation",
            "CloudWatchLogsLogGroupArn": "cloud_watch_logs_loggroup_arn",
            "CloudWatchLogsRoleArn": "cloud_watch_logs_role_arn",
            "KmsKeyId": "kms_key_id",
            "IsOrganizationTrail": "is_organization_trail",
        }
    )

    for raw_parameter, present_parameter in parameters.items():
        if raw_parameter in update_payload:
            result["ret"][present_parameter] = update_payload[raw_parameter]
            result["comment"] = result["comment"] + (
                f"Update {present_parameter}: {update_payload[raw_parameter]}",
            )
    return result


async def update_trail_attributes(
    hub,
    ctx,
    before: Dict[str, Any],
    resource_id: str,
    is_logging: bool,
    update_insight_selectors: List,
    update_event_selectors: List,
    update_advanced_event_selectors: List,
) -> Dict[str, Any]:
    result = dict(comment=(), result=True, ret=None)
    exising_attributes = await existing_attributes(hub, ctx, before, resource_id)
    result["ret"] = {}
    if not exising_attributes["result"]:
        result["comment"] = exising_attributes["comment"]
        result["result"] = False
    if (
        is_logging is not None
        and exising_attributes["result"]
        and is_logging != exising_attributes["ret"].get("existingTrailStatus")
    ):
        if not ctx.get("test", False):
            if is_logging:
                update_ret = await hub.exec.boto3.client.cloudtrail.start_logging(
                    ctx, Name=before["TrailARN"]
                )
            else:
                update_ret = await hub.exec.boto3.client.cloudtrail.stop_logging(
                    ctx, Name=before["TrailARN"]
                )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
                return result
        result["ret"]["is_logging"] = is_logging
        result["comment"] = result["comment"] + (f"Update is_logging: {is_logging}",)
    if (
        exising_attributes["result"]
        and update_insight_selectors is not None
        and (
            not hub.tool.aws.state_comparison_utils.are_lists_identical(
                update_insight_selectors,
                exising_attributes["ret"].get("existingInsightSelectors"),
            )
        )
    ):
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.cloudtrail.put_insight_selectors(
                ctx,
                TrailName=before["TrailARN"],
                InsightSelectors=update_insight_selectors,
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
                return result
        result["ret"]["insight_selectors"] = update_insight_selectors
        result["comment"] = result["comment"] + (
            f"Update insight_selectors: {update_insight_selectors}",
        )

    ## Update event selectors
    if (
        exising_attributes["result"]
        and update_event_selectors is not None
        and (
            not hub.tool.aws.state_comparison_utils.are_lists_identical(
                update_event_selectors, exising_attributes["ret"].get("event_selectors")
            )
        )
    ):
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.cloudtrail.put_event_selectors(
                ctx,
                TrailName=before["TrailARN"],
                EventSelectors=update_event_selectors,
                AdvancedEventSelectors=update_advanced_event_selectors,
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
                return result
        result["ret"]["event_selectors"] = update_event_selectors
        result["comment"] = result["comment"] + (
            f"Update event_selectors: {update_event_selectors}",
        )

    if (
        exising_attributes["result"]
        and update_advanced_event_selectors is not None
        and (
            not hub.tool.aws.state_comparison_utils.are_lists_identical(
                update_advanced_event_selectors,
                exising_attributes["ret"].get("advanced_event_selectors"),
            )
        )
    ):
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.cloudtrail.put_event_selectors(
                ctx,
                TrailName=before["TrailARN"],
                EventSelectors=update_event_selectors,
                AdvancedEventSelectors=update_advanced_event_selectors,
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
                return result
        result["ret"]["advanced_event_selectors"] = update_advanced_event_selectors
        result["comment"] = result["comment"] + (
            f"Update advanced_event_selectors: {update_advanced_event_selectors}",
        )

    return result


async def existing_attributes(
    hub, ctx, before: Dict[str, Any], resource_id: str
) -> Dict[str, Any]:
    result = dict(comment=(), result=True, ret=None)
    existing_logging = None
    existing_insight_selectors = None
    existing_event_selectors = None
    existing_advanced_event_selectors = None
    if not ctx.get("test", False):
        existingTrailStatus = await hub.exec.boto3.client.cloudtrail.get_trail_status(
            ctx, Name=resource_id
        )
        if existingTrailStatus["result"]:
            existing_logging = existingTrailStatus["ret"]["IsLogging"]
        else:
            result["comment"] = result["comment"] + existingTrailStatus["comment"]
            result["result"] = False

    if not ctx.get("test", False):
        existingInsightSelectors = (
            await hub.exec.boto3.client.cloudtrail.get_insight_selectors(
                ctx,
                TrailName=resource_id,
            )
        )
        if before["HasInsightSelectors"]:
            if existingInsightSelectors["result"]:
                existing_insight_selectors = existingInsightSelectors["ret"][
                    "InsightSelectors"
                ]
            else:
                result["comment"] = (
                    result["comment"] + existingInsightSelectors["comment"]
                )
                result["result"] = False

    if not ctx.get("test", False):
        existingEventSelectors = (
            await hub.exec.boto3.client.cloudtrail.get_event_selectors(
                ctx,
                TrailName=resource_id,
            )
        )
        if existingEventSelectors["result"]:
            if "EventSelectors" in existingEventSelectors["ret"]:
                existing_event_selectors = existingEventSelectors["ret"][
                    "EventSelectors"
                ]
            else:
                existing_advanced_event_selectors = existingEventSelectors["ret"][
                    "AdvancedEventSelectors"
                ]
        else:
            result["comment"] = result["comment"] + existingEventSelectors["comment"]
            result["result"] = False

    result["ret"] = {
        "existingTrailStatus": existing_logging,
        "existingInsightSelectors": existing_insight_selectors,
        "event_selectors": existing_event_selectors,
        "advanced_event_selectors": existing_advanced_event_selectors,
    }
    return result



"""State module for managing Amazon Application Autoscaling scaling policies."""
import copy
import re
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.application_autoscaling.scaling_target.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    policy_name: str,
    service_namespace: str,
    scaling_resource_id: str,
    scalable_dimension: str,
    policy_type: str,
    resource_id: str = None,
    step_scaling_policy_configuration: make_dataclass(
        "StepScalingPolicyConfiguration",
        [
            ("AdjustmentType", str, field(default=None)),
            (
                "StepAdjustments",
                List[
                    make_dataclass(
                        "StepAdjustment",
                        [
                            ("MetricIntervalLowerBound", float, field(default=None)),
                            ("MetricIntervalUpperBound", float, field(default=None)),
                            ("ScalingAdjustment", int, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            ("MinAdjustmentMagnitude", int, field(default=None)),
            ("Cooldown", int, field(default=None)),
            ("MetricAggregationType", str, field(default=None)),
        ],
    ) = None,
    target_tracking_scaling_policy_configuration: make_dataclass(
        "TargetTrackingScalingPolicyConfiguration",
        [
            ("TargetValue", float),
            (
                "PredefinedMetricSpecification",
                make_dataclass(
                    "PredefinedMetricSpecification",
                    [
                        ("PredefinedMetricType", str),
                        ("ResourceLabel", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "CustomizedMetricSpecification",
                make_dataclass(
                    "CustomizedMetricSpecification",
                    [
                        ("MetricName", str),
                        ("Namespace", str),
                        ("Statistic", str),
                        (
                            "Dimensions",
                            List[
                                make_dataclass(
                                    "MetricDimension", [("Name", str), ("Value", str)]
                                )
                            ],
                            field(default=None),
                        ),
                        ("Unit", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("ScaleOutCooldown", int, field(default=None)),
            ("ScaleInCooldown", int, field(default=None)),
            ("DisableScaleIn", bool, field(default=None)),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates or updates a scaling policy for an Application Auto Scaling scalable target.

    Each scalable target is identified by a service namespace, resource ID, and scalable dimension.
    A scaling policy applies to the scalable target identified by those three attributes.
    You cannot create a scaling policy until you have registered the resource as a scalable target.

    Multiple scaling policies can be in force at the same time for the same scalable
    target. You can have one or more target tracking scaling policies, one or more step scaling policies, or both.
    However, there is a chance that multiple policies could conflict, instructing the scalable target to scale out
    or in at the same time. Application Auto Scaling gives precedence to the policy that provides the largest
    capacity for both scale out and scale in. For example, if one policy increases capacity by 3, another policy
    increases capacity by 200 percent, and the current capacity is 10, Application Auto Scaling uses the policy with
    the highest calculated capacity (200% of 10 = 20) and scales out to 30.

    We recommend caution, however, when using target tracking scaling policies with step scaling policies because
    conflicts between these policies can cause undesirable behavior.
    For example, if the step scaling policy initiates a scale-in activity before the
    target tracking policy is ready to scale in, the scale-in activity will not be blocked. After the scale-in
    activity completes, the target tracking policy could instruct the scalable target to scale out again.

    For more information, see Target tracking scaling policies and Step scaling policies in
    the Application Auto Scaling User Guide.

    .. Note::
        If a scalable target is deregistered, the scalable target is no longer available to execute scaling
        policies. Any scaling policies that were specified for the scalable target are deleted.

    Args:
        name(str):
            An Idem name of the resource.

        policy_name(str):
            The name of the scaling policy.

        service_namespace(str):
            The namespace of the Amazon Web Services service that provides the resource.
            For a resource provided by your own application or service, use custom-resource instead.

        scaling_resource_id(str): The identifier of the resource associated with the scaling policy.
        This string consists of the resource type and unique identifier.

            * ECS service - The resource type is service and the unique identifier is the cluster name and service name.
              Example: service/default/sample-webapp.
            * Spot Fleet - The resource type is spot-fleet-request and the unique identifier is the Spot Fleet request
              ID. Example: spot-fleet-request/sfr-73fbd2ce-aa30-494c-8788-1cee4EXAMPLE.
            * EMR cluster - The resource type is instancegroup and the unique identifier is the cluster ID and instance
              group ID. Example: instancegroup/j-2EEZNYKUA1NTV/ig-1791Y4E1L8YI0.
            * AppStream 2.0 fleet - The resource type is fleet and the unique identifier is the fleet name.
              Example: fleet/sample-fleet.
            * DynamoDB table - The resource type is table and the unique identifier is the table name.
              Example: table/my-table.
            * DynamoDB global secondary index - The resource type is index and the unique identifier is the index name.
              Example: table/my-table/index/my-table-index.
            * Aurora DB cluster - The resource type is cluster and the unique identifier is the cluster name.
              Example:cluster:my-db-cluster.
            * SageMaker endpoint variant - The resource type is variant and the unique identifier is the resource ID.
              Example: endpoint/my-end-point/variant/KMeansClustering.
            * Custom resources are not supported with a resource type. This parameter must specify the OutputValue
              from the CloudFormation template stack used to access the resources. The unique
              identifier is defined by the service provider. More information is available in our GitHub repository.
            * Amazon Comprehend document classification endpoint - The resource type and unique identifier are specified
              using the endpoint ARN.
              Example: arn:aws:comprehend:us-west-2:123456789012:document-classifier-endpoint/EXAMPLE.
            * Amazon Comprehend entity recognizer endpoint - The resource type and unique identifier are specified using
              the endpoint ARN. Example: arn:aws:comprehend:us-west-2:123456789012:entity-recognizer-endpoint/EXAMPLE.
            * Lambda provisioned concurrency - The resource type is function and the unique identifier is the function
              name with a function version or alias name suffix that is not $LATEST.
              Example: function:my-function:prod or function:my-function:1.
            * Amazon Keyspaces table - The resource type is table and the unique identifier is the table name.
              Example: keyspace/mykeyspace/table/mytable.
            * Amazon MSK cluster - The resource type and unique identifier are specified using the cluster ARN. Example:
              arn:aws:kafka:us-east-1:123456789012:cluster/demo-cluster-1/6357e0b2-0e6a-4b86-a0b4-70df934c2e31-5.
            * Amazon ElastiCache replication group - The resource type is replication-group and the unique identifier
              is the replication group name. Example: replication-group/mycluster.
            * Neptune cluster - The resource type is cluster and the unique identifier is the cluster name.
              Example: cluster:mycluster.

        scalable_dimension(str):
            The scalable dimension. This string consists of the service namespace, resource type, and scaling property.

            * ecs:service:DesiredCount - The desired task count of an ECS service.
            * elasticmapreduce:instancegroup:InstanceCount - The instance count of an EMR Instance Group.
            * ec2:spot-fleet-request:TargetCapacity - The target capacity of a Spot Fleet.
            * appstream:fleet:DesiredCapacity - The desired capacity of an AppStream 2.0 fleet.
            * dynamodb:table:ReadCapacityUnits - The provisioned read capacity for a DynamoDB table.
            * dynamodb:table:WriteCapacityUnits - The provisioned write capacity for a DynamoDB table.
            * dynamodb:index:ReadCapacityUnits - The provisioned read capacity for a DynamoDB global secondary index.
            * dynamodb:index:WriteCapacityUnits - The provisioned write capacity for a DynamoDB global secondary index.
            * rds:cluster:ReadReplicaCount - The count of Aurora Replicas in an Aurora DB cluster.
              Available for Aurora MySQL-compatible edition and Aurora PostgreSQL- compatible edition.
            * sagemaker:variant:DesiredInstanceCount -
              The number of EC2 instances for an SageMaker model endpoint variant.
            * custom-resource:ResourceType:Property -
              The scalable dimension for a custom resource provided by your own application or service.
            * comprehend:document-classifier-endpoint:DesiredInferenceUnits -
              The number of inference units for an Amazon Comprehend document classification endpoint.
            * comprehend:entity-recognizer-endpoint:DesiredInferenceUnits -
              The number of inference units for an Amazon Comprehend entity recognizer endpoint.
            * lambda:function:ProvisionedConcurrency - The provisioned concurrency for a Lambda function.
            * cassandra:table:ReadCapacityUnits - The provisioned read capacity for an Amazon Keyspaces table.
            * cassandra:table:WriteCapacityUnits - The provisioned write capacity for an Amazon Keyspaces table.
            * kafka:broker-storage:VolumeSize -
              The provisioned volume size (in GiB) for brokers in an Amazon MSK cluster.
            * elasticache:replication-group:NodeGroups -
              The number of node groups for an Amazon ElastiCache replication group.
            * elasticache:replication-group:Replicas -
              The number of replicas per node group for an Amazon ElastiCache replication group.
            * neptune:cluster:ReadReplicaCount -
              The count of read replicas in an Amazon Neptune DB cluster.

        policy_type(str):
            The policy type. This parameter is required if you are creating a scaling policy.

            The following policy types are supported:

            * TargetTrackingScaling — Not supported for Amazon EMR
            * StepScaling — Not supported for DynamoDB, Amazon Comprehend, Lambda, Amazon Keyspaces, Amazon MSK,
              Amazon ElastiCache, or Neptune.

            For more information, see Target tracking scaling policies and Step scaling policies in
            the Application Auto Scaling User Guide. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        step_scaling_policy_configuration(dict[str, Any], Optional):
            A step scaling policy.
            This parameter is required if you are creating a policy and the policy type is StepScaling.
            Defaults to None.

            * AdjustmentType (str, Optional):
                Specifies how the ScalingAdjustment value in a StepAdjustment is interpreted (for example, an
                absolute number or a percentage). The valid values are ChangeInCapacity, ExactCapacity, and
                PercentChangeInCapacity.
                AdjustmentType is required if you are adding a new step scaling policy configuration.
            * StepAdjustments (list[dict[str, Any]], Optional):
                A set of adjustments that enable you to scale based on the size of the alarm breach. At least
                one step adjustment is required if you are adding a new step scaling policy configuration.

                * MetricIntervalLowerBound (float, Optional):
                    The lower bound for the difference between the alarm threshold and the CloudWatch metric. If the
                    metric value is above the breach threshold, the lower bound is inclusive (the metric must be
                    greater than or equal to the threshold plus the lower bound). Otherwise, it is exclusive (the
                    metric must be greater than the threshold plus the lower bound). A null value indicates negative
                    infinity.
                * MetricIntervalUpperBound (float, Optional):
                    The upper bound for the difference between the alarm threshold and the CloudWatch metric. If the
                    metric value is above the breach threshold, the upper bound is exclusive (the metric must be
                    less than the threshold plus the upper bound). Otherwise, it is inclusive (the metric must be
                    less than or equal to the threshold plus the upper bound). A null value indicates positive
                    infinity. The upper bound must be greater than the lower bound.
                * ScalingAdjustment (int):
                    The amount by which to scale, based on the specified adjustment type. A positive value adds to
                    the current capacity while a negative number removes from the current capacity. For exact capacity,
                    you must specify a positive value.

            * MinAdjustmentMagnitude (int, Optional):
                The minimum value to scale by when the adjustment type is PercentChangeInCapacity. For example,
                suppose that you create a step scaling policy to scale out an Amazon ECS service by 25 percent
                and you specify a MinAdjustmentMagnitude of 2. If the service has 4 tasks and the scaling policy
                is performed, 25 percent of 4 is 1. However, because you specified a MinAdjustmentMagnitude of
                2, Application Auto Scaling scales out the service by 2 tasks.

            * Cooldown (int, Optional):
                The amount of time, in seconds, to wait for a previous scaling activity to take effect.  With
                scale-out policies, the intention is to continuously (but not excessively) scale out. After
                Application Auto Scaling successfully scales out using a step scaling policy, it starts to
                calculate the cooldown time. The scaling policy won't increase the desired capacity again unless
                either a larger scale out is triggered or the cooldown period ends. While the cooldown period is
                in effect, capacity added by the initiating scale-out activity is calculated as part of the
                desired capacity for the next scale-out activity. For example, when an alarm triggers a step
                scaling policy to increase the capacity by 2, the scaling activity completes successfully, and a
                cooldown period starts. If the alarm triggers again during the cooldown period but at a more
                aggressive step adjustment of 3, the previous increase of 2 is considered part of the current
                capacity. Therefore, only 1 is added to the capacity. With scale-in policies, the intention is
                to scale in conservatively to protect your application’s availability, so scale-in activities
                are blocked until the cooldown period has expired. However, if another alarm triggers a scale-
                out activity during the cooldown period after a scale-in activity, Application Auto Scaling
                scales out the target immediately. In this case, the cooldown period for the scale-in activity
                stops and doesn't complete.

                Application Auto Scaling provides a default value of 600 for Amazon ElastiCache replication groups and a default value of 300 for the following scalable targets:
                 * AppStream 2.0 fleets
                 * Aurora DB clusters
                 * ECS services
                 * EMR clusters
                 * Neptune clusters
                 * SageMaker endpoint variants
                 * Spot Fleets
                 * Custom resources

                For all other scalable targets, the default value is 0:
                 * Amazon Comprehend document classification and entity recognizer endpoints
                 * DynamoDB tables and global secondary indexes
                 * Amazon Keyspaces tables
                 * Lambda provisioned concurrency
                 * Amazon MSK broker storage

            * MetricAggregationType (str, Optional):
                The aggregation type for the CloudWatch metrics. Valid values are Minimum, Maximum, and Average.
                If the aggregation type is null, the value is treated as Average.

        target_tracking_scaling_policy_configuration(dict[str, Any], Optional):
            A target tracking scaling policy. Includes support for predefined or customized metrics.
            This parameter is required if you are creating a policy and the policy type is TargetTrackingScaling.
            Defaults to None.

            * TargetValue (float):
                The target value for the metric. Although this property accepts numbers of type Double, it won't
                accept values that are either too small or too large. Values must be in the range of -2^360 to
                2^360. The value must be a valid number based on the choice of metric. For example, if the
                metric is CPU utilization, then the target value is a percent value that represents how much of
                the CPU can be used before scaling out.
            * PredefinedMetricSpecification (dict[str, Any], Optional):
                A predefined metric. You can specify either a predefined metric or a customized metric.

                * PredefinedMetricType (str):
                    The metric type. The ALBRequestCountPerTarget metric type applies only to Spot Fleet requests and
                    ECS services.
                * ResourceLabel (str, Optional):
                    Identifies the resource associated with the metric type. You can't specify a resource label
                    unless the metric type is ALBRequestCountPerTarget and there is a target group attached to the
                    Spot Fleet request or ECS service.

                    You create the resource label by appending the final portion
                    of the load balancer ARN and the final portion of the target group ARN into a single value,
                    separated by a forward slash (/).

                    The format of the resource label is:
                    app/my-alb/778d41231b141a0f/targetgroup/my-alb-target-group/943f017f100becff.

                    Where:

                    * app/<load-balancer-name>/<load-balancer-id> is the final portion of the load balancer ARN
                    * targetgroup/<target-group-name>/<target-group-id> is the final portion of the target group ARN.

                    To find the ARN for an Application Load Balancer, use the DescribeLoadBalancers API operation.
                    To find the ARN for the target group, use the DescribeTargetGroups API operation.

            * CustomizedMetricSpecification (dict[str, Any], Optional):
                A customized metric. You can specify either a predefined metric or a customized metric.

                * MetricName (str): The name of the metric.
                * Namespace (str): The namespace of the metric.
                * Dimensions (List[Dict[str, Any]], Optional): The dimensions of the metric.

                  Conditional: If you published your metric with dimensions, you must specify the same dimensions in
                  your scaling policy.

                  * Name (str): The name of the dimension.
                  * Value (str): The value of the dimension.

                * Statistic (str): The statistic of the metric.
                * Unit (str, Optional): The unit of the metric.

            * ScaleOutCooldown (int, Optional):
                The amount of time, in seconds, to wait for a previous scale-out activity to take effect.

                With the scale-out cooldown period, the intention is to continuously (but not excessively) scale out.
                After Application Auto Scaling successfully scales out using a target tracking scaling policy,
                it starts to calculate the cooldown time. The scaling policy won't increase the desired capacity
                again unless either a larger scale out is triggered or the cooldown period ends. While the
                cooldown period is in effect, the capacity added by the initiating scale-out activity is
                calculated as part of the desired capacity for the next scale-out activity.

                Application Auto Scaling provides a default value of 600 for Amazon ElastiCache replication groups and a default value of 300 for the following scalable targets:
                 * AppStream 2.0 fleets
                 * Aurora DB clusters
                 * ECS services
                 * EMR clusters
                 * Neptune clusters
                 * SageMaker endpoint variants
                 * Spot Fleets
                 * Custom resources

                For all other scalable targets, the default value is 0:
                 * Amazon Comprehend document classification and entity recognizer endpoints
                 * DynamoDB tables and global secondary indexes
                 * Amazon Keyspaces tables
                 * Lambda provisioned concurrency
                 * Amazon MSK broker storage

            * ScaleInCooldown (int, Optional):
                The amount of time, in seconds, after a scale-in activity completes before another scale-in
                activity can start.

                With the scale-in cooldown period, the intention is to scale in conservatively to protect your
                application’s availability, so scale-in activities are blocked
                until the cooldown period has expired. However, if another alarm triggers a scale-out activity
                during the scale-in cooldown period, Application Auto Scaling scales out the target immediately.
                In this case, the scale-in cooldown period stops and doesn't complete.

                Application Auto Scaling provides a default value of 600 for Amazon ElastiCache replication groups  and a default value of 300 for the following scalable targets:
                 * AppStream 2.0 fleets
                 * Aurora DB clusters
                 * ECS services
                 * EMR clusters
                 * Neptune clusters
                 * SageMaker endpoint variants
                 * Spot Fleets
                 * Custom resources

                For all other scalable targets, the default value is 0:
                 * Amazon Comprehend document classification and entity recognizer endpoints
                 * DynamoDB tables and global secondary indexes
                 * Amazon Keyspaces tables
                 * Lambda provisioned concurrency
                 * Amazon MSK broker storage

            * DisableScaleIn (bool, Optional):
                Indicates whether scale in by the target tracking scaling policy is disabled. If the value is
                true, scale in is disabled and the target tracking scaling policy won't remove capacity from the
                scalable target. Otherwise, scale in is enabled and the target tracking scaling policy can
                remove capacity from the scalable target. The default value is false.

    Request Syntax:
        .. code-block:: sls

            [scaling_policy_name]:
              aws.application_autoscaling.scaling_policy.present:
                - name: 'string'
                - policy_name: 'string'
                - service_namespace: 'string'
                - scalable_dimension: 'string'
                - scaling_resource_id: 'string'
                - policy_type: 'string'
                - step_scaling_policy_configuration:
                    AdjustmentType: 'string'
                    StepAdjustments:
                        - MetricIntervalLowerBound: 'int'
                          MetricIntervalUpperBound: 'int'
                          ScalingAdjustment: 'int'
                        - MetricIntervalLowerBound: 'int'
                          MetricIntervalUpperBound: 'int'
                          ScalingAdjustment: 'int'
                        - MetricIntervalLowerBound: 'int'
                          ScalingAdjustment: 'int'
                    MinAdjustmentMagnitude: 'int'
                    Cooldown: 'int'
                    MetricAggregationType: 'string'


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            rds_scaling_policy_name:
              aws.application_autoscaling.scaling_policy.present:
                - name: rds_scaling_policy
                - policy_name: rds_scaling_policy
                - service_namespace: rds
                - scalable_dimension: rds:cluster:ReadReplicaCount
                - scaling_resource_id: idem-test-rds-aurora-table
                - policy_type: StepScaling
                - step_scaling_policy_configuration:
                    AdjustmentType: PercentChangeInCapacity
                    StepAdjustments:
                        - MetricIntervalLowerBound: 0
                          MetricIntervalUpperBound: 15
                          ScalingAdjustment: 1
                        - MetricIntervalLowerBound: 15
                          MetricIntervalUpperBound: 25
                          ScalingAdjustment: 2
                        - MetricIntervalLowerBound: 25
                          ScalingAdjustment: 3
                    MinAdjustmentMagnitude: 1
                    Cooldown: 20
                    MetricAggregationType: Average
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    plan_state = None
    if resource_id:
        if not re.search(
            f"^({service_namespace})/({scaling_resource_id})/({scalable_dimension})/({policy_name})$",
            resource_id,
        ):
            result["comment"] = (
                f"Incorrect aws.application_autoscaling.scaling_policy resource_id: {resource_id}. Expected id {service_namespace}/{scaling_resource_id}/{scalable_dimension}/{policy_name}",
            )
            result["result"] = False
            return result
        before = await hub.exec.aws.application_autoscaling.scaling_policy.get(
            ctx=ctx,
            name=name,
            service_namespace=service_namespace,
            scaling_resource_id=scaling_resource_id,
            policy_names=[policy_name],
            scalable_dimension=scalable_dimension,
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        update_ret = await hub.tool.aws.application_autoscaling.scaling_policy.update_scaling_policy(
            ctx=ctx,
            name=name,
            before=result["old_state"],
            policy_name=policy_name,
            service_namespace=service_namespace,
            scaling_resource_id=scaling_resource_id,
            scalable_dimension=scalable_dimension,
            policy_type=policy_type,
            step_scaling_policy_configuration=step_scaling_policy_configuration,
            target_tracking_scaling_policy_configuration=target_tracking_scaling_policy_configuration,
        )
        result["comment"] = update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            for modified_param in update_ret["ret"]:
                plan_state[modified_param] = update_ret["ret"][modified_param]
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.application_autoscaling.scaling_policy", name=name
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "policy_name": policy_name,
                    "service_namespace": service_namespace,
                    "scaling_resource_id": scaling_resource_id,
                    "scalable_dimension": scalable_dimension,
                    "policy_type": policy_type,
                    "step_scaling_policy_configuration": step_scaling_policy_configuration,
                    "target_tracking_scaling_policy_configuration": target_tracking_scaling_policy_configuration,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.application_autoscaling.scaling_policy", name=name
            )
            return result
        ret = await hub.exec.boto3.client["application-autoscaling"].put_scaling_policy(
            ctx,
            PolicyName=policy_name,
            ServiceNamespace=service_namespace,
            ResourceId=scaling_resource_id,
            ScalableDimension=scalable_dimension,
            PolicyType=policy_type,
            StepScalingPolicyConfiguration=step_scaling_policy_configuration,
            TargetTrackingScalingPolicyConfiguration=target_tracking_scaling_policy_configuration,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.application_autoscaling.scaling_policy", name=name
        )
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.aws.application_autoscaling.scaling_policy.get(
                ctx=ctx,
                name=name,
                service_namespace=service_namespace,
                scaling_resource_id=scaling_resource_id,
                policy_names=[policy_name],
                scalable_dimension=scalable_dimension,
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] += after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    scaling_resource_id: str,
    policy_name: str,
    service_namespace: str,
    scalable_dimension: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes the specified scaling policy for an Application Auto Scaling scalable target.

    Deleting a step scaling policy deletes the underlying alarm action, but does not delete the CloudWatch alarm
    associated with the scaling policy, even if it no longer has an associated action. For more information, see
    Delete a step scaling policy and Delete a target tracking scaling policy in the Application Auto Scaling User Guide.

    Args:
        name(str):
            An Idem name of the resource.

        policy_name(str):
            The name of the scaling policy.

        service_namespace(str):
            The namespace of the Amazon Web Services service that provides the resource.
            For a resource provided by your own application or service, use custom-resource instead.

        scaling_resource_id(str):
            The identifier of the resource associated with the scalable target.
            This string consists of the resource type and unique identifier.

            * ECS service - The resource type is service and the unique identifier is the cluster name and service name.
              Example: service/default/sample-webapp.
            * Spot Fleet -
              The resource type is spot-fleet-request and the unique identifier is the Spot Fleet request ID.
              Example: spot-fleet-request/sfr-73fbd2ce-aa30-494c-8788-1cee4EXAMPLE.
            * EMR cluster - The resource type is instancegroup and the unique identifier is the cluster ID and instance
              group ID. Example: instancegroup/j-2EEZNYKUA1NTV/ig-1791Y4E1L8YI0.
            * AppStream 2.0 fleet - The resource type is fleet and the unique identifier is the fleet name.
              Example: fleet/sample-fleet.
            * DynamoDB table - The resource type is table and the unique identifier is the table name.
              Example: table/my-table.
            * DynamoDB global secondary index - The resource type is index and the
              unique identifier is the index name. Example: table/my-table/index/my-table-index.
            * Aurora DB cluster - The resource type is cluster and the unique identifier is the cluster name.
              Example: cluster:my-db-cluster.
            * SageMaker endpoint variant - The resource type is variant and the
              unique identifier is the resource ID. Example: endpoint/my-end-point/variant/KMeansClustering.
            * Custom resources are not supported with a resource type. This parameter must specify the
              OutputValue from the CloudFormation template stack used to access the resources. The unique
              identifier is defined by the service provider. More information is available in our GitHub
              repository.
            * Amazon Comprehend document classification endpoint -
              The resource type and unique identifier are specified using the endpoint ARN.
              Example: arn:aws:comprehend:us-west-2:123456789012:document-classifier-endpoint/EXAMPLE.
            * Amazon Comprehend entity recognizer endpoint -
              The resource type and unique identifier are specified using the endpoint ARN.
              Example: arn:aws:comprehend:us-west-2:123456789012:entity-recognizer-endpoint/EXAMPLE.
            * Lambda provisioned concurrency - The resource type is function and the unique identifier is the
              function name with a function version or alias name suffix that is not $LATEST. Example:
              function:my-function:prod or function:my-function:1.
            * Amazon Keyspaces table -
              The resource type is table and the unique identifier is the table name. Example:
              keyspace/mykeyspace/table/mytable.
            * Amazon MSK cluster -
              The resource type and unique identifier are specified using the cluster ARN.
              Example: arn:aws:kafka:us-east-1:123456789012:cluster/demo-cluster-1/6357e0b2-0e6a-4b86-a0b4-70df934c2e31-5.
            * Amazon ElastiCache replication group - The resource type is replication-group and the unique identifier
              is the replication group name. Example: replication-group/mycluster.
            * Neptune cluster -
              The resource type is cluster and the unique identifier is the cluster name. Example: cluster:mycluster.

        scalable_dimension(str):
            The scalable dimension. This string consists of the service namespace, resource type, and scaling property.

            * ecs:service:DesiredCount - The desired task count of an ECS service.
            * elasticmapreduce:instancegroup:InstanceCount - The instance count of an EMR Instance Group.
            * ec2:spot-fleet-request:TargetCapacity - The target capacity of a Spot Fleet.
            * appstream:fleet:DesiredCapacity - The desired capacity of an AppStream 2.0 fleet.
            * dynamodb:table:ReadCapacityUnits - The provisioned read capacity for a DynamoDB table.
            * dynamodb:table:WriteCapacityUnits - The provisioned write capacity for a DynamoDB table.
            * dynamodb:index:ReadCapacityUnits - The provisioned read capacity for a DynamoDB global secondary index.
            * dynamodb:index:WriteCapacityUnits - The provisioned write capacity for a DynamoDB global secondary index.
            * rds:cluster:ReadReplicaCount - The count of Aurora Replicas in an Aurora DB cluster.
              Available for Aurora MySQL-compatible edition and Aurora PostgreSQL- compatible edition.
            * sagemaker:variant:DesiredInstanceCount -
              The number of EC2 instances for an SageMaker model endpoint variant.
            * custom-resource:ResourceType:Property -
              The scalable dimension for a custom resource provided by your own application or service.
            * comprehend:document-classifier-endpoint:DesiredInferenceUnits -
              The number of inference units for an Amazon Comprehend document classification endpoint.
            * comprehend:entity-recognizer- endpoint:DesiredInferenceUnits -
              The number of inference units for an Amazon Comprehend entity recognizer endpoint.
            * lambda:function:ProvisionedConcurrency - The provisioned concurrency for a Lambda function.
            * cassandra:table:ReadCapacityUnits - The provisioned read capacity for an Amazon Keyspaces table.
            * cassandra:table:WriteCapacityUnits - The provisioned write capacity for an Amazon Keyspaces table.
            * kafka:broker-storage:VolumeSize -
              The provisioned volume size(in GiB) for brokers in an Amazon MSK cluster.
            * elasticache:replication-group:NodeGroups -
              The number of node groups for an Amazon ElastiCache replication group.
            * elasticache:replication-group:Replicas -
              The number of replicas per node group for an Amazon ElastiCache replication group.
            * neptune:cluster:ReadReplicaCount - The count of read replicas in an Amazon Neptune DB cluster.

        resource_id(str, Optional):
            An identifier of the resource in the provider.

    Request Syntax:
        .. code-block:: sls

            [scaling_policy_name]:
              aws.application_autoscaling.scaling_policy.absent:
                - name: 'string'
                - policy_name: 'string'
                - service_namespace: 'string'
                - scalable_dimension: 'string'
                - scaling_resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            rds_scaling_policy_name:
              aws.application_autoscaling.scaling_policy.absent:
                - name: rds_scaling_policy
                - policy_name: rds_scaling_policy
                - service_namespace: rds
                - scalable_dimension: rds:cluster:ReadReplicaCount
                - scaling_resource_id: idem-test-rds-aurora-table
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.application_autoscaling.scaling_policy", name=name
        )
        return result
    if not re.search(
        f"^({service_namespace})/({scaling_resource_id})/({scalable_dimension})/({policy_name})$",
        resource_id,
    ):
        result["comment"] = (
            f"Incorrect aws.application_autoscaling.scaling_policy resource_id: {resource_id}. Expected id {service_namespace}/{scaling_resource_id}/{scalable_dimension}/{policy_name}",
        )
        result["result"] = False
        return result
    before = await hub.exec.aws.application_autoscaling.scaling_policy.get(
        ctx=ctx,
        name=name,
        service_namespace=service_namespace,
        scaling_resource_id=scaling_resource_id,
        policy_names=[policy_name],
        scalable_dimension=scalable_dimension,
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.application_autoscaling.scaling_policy", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.application_autoscaling.scaling_policy", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client[
            "application-autoscaling"
        ].delete_scaling_policy(
            ctx,
            PolicyName=policy_name,
            ServiceNamespace=service_namespace,
            ResourceId=scaling_resource_id,
            ScalableDimension=scalable_dimension,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = result[
            "comment"
        ] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.application_autoscaling.scaling_policy", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Describes the Application Auto Scaling scaling policies for the specified service namespace. You can filter the
    results using ResourceId, ScalableDimension, and PolicyNames. For more information, see Target tracking scaling
    policies and Step scaling policies in the Application Auto Scaling User Guide.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.application_autoscaling.scaling_policy
    """
    result = {}
    # service_name_spaces supported by AWS. loop through all the service name spaces and list policies of each service.
    service_name_spaces = [
        "ecs",
        "elasticmapreduce",
        "ec2",
        "appstream",
        "dynamodb",
        "rds",
        "sagemaker",
        "custom-resource",
        "comprehend",
        "lambda",
        "cassandra",
        "kafka",
        "elasticache",
        "neptune",
    ]
    for service_name_space in service_name_spaces:
        ret = await hub.exec.boto3.client[
            "application-autoscaling"
        ].describe_scaling_policies(ctx, ServiceNamespace=service_name_space)
        if not ret["result"]:
            hub.log.debug(
                f"Could not describe scaling_policies for service name space {service_name_space}. {ret['comment']}"
            )
            continue
        for scaling_policy in ret["ret"]["ScalingPolicies"]:
            resource_id = f"{scaling_policy.get('ServiceNamespace')}/{scaling_policy.get('ResourceId')}/{scaling_policy.get('ScalableDimension')}/{scaling_policy.get('PolicyName')}"
            resource_translated = hub.tool.aws.application_autoscaling.conversion_utils.convert_raw_scaling_policy_to_present(
                ctx, raw_resource=scaling_policy, idem_resource_name=resource_id
            )
            result[resource_id] = {
                "aws.application_autoscaling.scaling_policy.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }
    return result



"""Exec module for managing IAM Group Membership."""


async def get(hub, ctx, name: str, resource_id: str):
    """Get IAM Group membership.

    Provide group name as input.

    Args:
        hub: required for functions in hub.

        ctx: context.

        name(str): An Idem name of the IAM group.

        resource_id(str): IAM Group Name
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.iam.get_group(ctx, GroupName=resource_id)
    if not ret["result"]:
        if "NoSuchEntity" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.iam.group_membership", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["result"] = False
        result["comment"] += list(ret["comment"])
        return result

    users = ret["ret"].get("Users")

    result[
        "ret"
    ] = hub.tool.aws.iam.conversion_utils.convert_raw_group_membership_to_present(
        group=resource_id, users=[user.get("UserName") for user in users]
    )

    return result



"""State module for managing AWS S3 buckets."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    acl: str = None,
    create_bucket_configuration: make_dataclass(
        "CreateBucketConfiguration", [("LocationConstraint", str, field(default=None))]
    ) = None,
    grant_full_control: str = None,
    grant_read: str = None,
    grant_read_acp: str = None,
    grant_write: str = None,
    grant_write_acp: str = None,
    object_lock_enabled_for_bucket: bool = None,
    object_ownership: str = None,
    object_lock_configuration: make_dataclass(
        "ObjectLockConfiguration",
        [
            ("ObjectLockEnabled", str, field(default=None)),
            (
                "Rule",
                make_dataclass(
                    "ObjectLockRule",
                    [
                        (
                            "DefaultRetention",
                            make_dataclass(
                                "DefaultRetention",
                                [
                                    ("Mode", str, field(default=None)),
                                    ("Days", int, field(default=None)),
                                    ("Years", int, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        )
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    """Create an AWS S3 Bucket.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional): AWS S3 Bucket id

        acl(str, Optional):
            The associated acl for this bucket ('private', 'public-read', 'public-read-write', 'authenticated-read').

        create_bucket_configuration(Dict[str, Any], Optional):
            The configuration information for the bucket. Defaults to None.
            LocationConstraint (str, Optional): Specifies the Region where the bucket will be created.
            If you don't specify a Region, the bucket is created in the US East (N. Virginia) Region (us-east-1).

        grant_full_control (str, Optional):
            Allows grantee the read, write, read ACP, and write ACP permissions on the bucket.

        grant_read (str, Optional):
            Allows grantee to list the objects in the bucket.

        grant_read_acp (str, Optional):
            Allows grantee to read the bucket ACL.

        grant_write (str, Optional):
            Allows grantee to create new objects in the bucket.
            For the bucket and object owners of existing objects, also allows deletions and overwrites of those objects.

        grant_write_acp (str, Optional):
            Allows grantee to write the ACL for the applicable bucket.

        object_lock_enabled_for_bucket (bool, Optional):
            Specifies whether you want S3 Object Lock to be enabled for the new bucket.

        object_ownership (str, Optional):
            The container element for object ownership for a bucket's ownership controls.

            * BucketOwnerPreferred
                Objects uploaded to the bucket change ownership to the bucket owner if
                the objects are uploaded with the bucket-owner-full-control canned ACL.
            * ObjectWriter
                The uploading account will own the object if the object is uploaded
                with the bucket-owner-full-control canned ACL.
            * BucketOwnerEnforced
                Access control lists (ACLs) are disabled and no longer affect permissions.
                The bucket owner automatically owns and has full control over every object
                in the bucket. The bucket only accepts PUT requests that don't specify an
                ACL or bucket owner full control ACLs, such as the bucket-owner-full-control
                canned ACL or an equivalent form of this ACL expressed in the XML format.

        object_lock_configuration (dict, Optional):
            The Object Lock configuration that you want to apply to the specified bucket.

            * ObjectLockEnabled (str):
                Indicates whether this bucket has an Object Lock configuration enabled. Enable ObjectLockEnabled
                when you apply ObjectLockConfiguration to a bucket.
            * Rule (dict):
                Specifies the Object Lock rule for the specified object. Enable this rule when you apply ObjectLockConfiguration
                to a bucket. Bucket settings require both a mode and a period. The period can be either Days or Years but you must select one.
                You cannot specify Days and Years at the same time.

                * DefaultRetention (dict):
                    The default Object Lock retention mode and period that you want to apply to new objects placed
                    in the specified bucket. Bucket settings require both a mode and a period. The period can be either Days or Years
                    but you must select one. You cannot specify Days and Years at the same time.

                    * Mode (str):
                        The default Object Lock retention mode you want to apply to new objects placed in the specified bucket.
                        Must be used with either Days or Years .
                    * Days (int):
                        The number of days that you want to specify for the default retention period. Must be used with Mode .
                    * Years (int):
                        The number of years that you want to specify for the default retention period. Must be used with Mode .

        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"Key": tag-key, "Value": tag-value}]`` to associate with the AMI.

            * Key (str):
                The key name that can be used to look up or retrieve the associated value. For example,
                Department or Cost Center are common choices.

            * Value (str):
                The value associated with this tag. For example, tags with a key name of Department could have
                values such as Human Resources, Accounting, and Support. Tags with a key name of Cost Center
                might have values that consist of the number associated with the different cost centers in your
                company. Typically, many resources have tags with the same key name but with different values.
                Amazon Web Services always interprets the tag Value as a single string. If you need to store an
                array, you can store comma-separated values in the string. However, you must interpret the value
                in your code.

    Request Syntax:
       .. code-block:: sls

            [s3-resource-id]:
              aws.s3.bucket.present:
                - name: "string"
                - acl: "private|public-read|public-read-write|authenticated-read"
                - create_bucket_configuration:
                    LocationConstraint: "string"
                - grant_full_control: "string"
                - grant_read: "string"
                - grant_read_acp: "string"
                - grant_write: "string"
                - grant_write_acp: "string"
                - object_lock_enabled_for_bucket: True|False
                - object_ownership: "BucketOwnerPreferred|ObjectWriter|BucketOwnerEnforced"
                - object_lock_configuration:
                    ObjectLockEnabled: "string"
                    Rule:
                      DefaultRetention:
                        Mode: "string"
                        Days: integer
                - tags:
                    - Key: "string"
                      Value: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            test_bucket-bb7bb32e9533:
              aws.s3.bucket.present:
                - name: "test_bucket-bb7bb32e9533"
                - acl: "private"
                - create_bucket_configuration:
                    LocationConstraint: "sa-east-1"
                - object_lock_enabled_for_bucket: True
                - object_lock_configuration:
                    ObjectLockEnabled: "Enabled"
                    Rule:
                      DefaultRetention:
                        Mode: "GOVERNANCE"
                        Days: 1
                - object_ownership: "BucketOwnerEnforced"
                - tags:
                    - Key: "Name1"
                      Value: "s3-test1"
                    - Key: "Name2"
                      Value: "s3-test2"


    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    tags_list = None
    tags_dict = None
    if tags is not None:
        if isinstance(tags, Dict):
            tags_list = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            tags_dict = tags
        else:
            tags_list = tags
            tags_dict = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if resource_id:
        resource = hub.tool.boto3.resource.create(ctx, "s3", "Bucket", resource_id)
        before = await hub.tool.boto3.resource.describe(resource)
    if before:
        convert_ret = await hub.exec.aws.s3.bucket.get(
            ctx, name=name, resource_id=resource_id
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
            return result
        result["old_state"] = copy.deepcopy(convert_ret["ret"].get("ret"))

        plan_state = copy.deepcopy(result["old_state"])
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            "aws.s3.bucket", name
        )
        # S3 Object Lock can only be enabled on bucket creation and cannot be enabled on existing buckets.
        # Changed the condition to fix the object_lock_enabled_for_bucket issue. The issue was when we provide
        # object_lock_enabled_for_bucket as false in sls it was creating changes constantly leading to
        # put_object_lock_configuration retry max out exception. The fix is updating the lock configuration only if
        # there is change in rules and does not try to update based on object_lock_enabled_for_bucket flag.
        if (
            object_lock_configuration is not None
            and object_lock_configuration
            != result["old_state"].get("object_lock_configuration", {})
        ):
            if not ctx.get("test", False):
                update_ret = (
                    await hub.exec.boto3.client.s3.put_object_lock_configuration(
                        ctx,
                        Bucket=resource_id,
                        ObjectLockConfiguration=object_lock_configuration,
                    )
                )
                if not update_ret["result"]:
                    result["comment"] = update_ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
            else:
                plan_state[
                    "object_lock_enabled_for_bucket"
                ] = object_lock_enabled_for_bucket
                plan_state["object_lock_configuration"] = object_lock_configuration
        if (
            tags_dict is not None
            and not hub.tool.aws.state_comparison_utils.compare_dicts(
                result["old_state"].get("tags"), tags_dict
            )
        ):
            if not ctx.get("test", False):
                update_ret = await hub.exec.boto3.client.s3.put_bucket_tagging(
                    ctx,
                    Bucket=resource_id,
                    Tagging={"TagSet": tags_list},
                )
                result["comment"] = result["comment"] + (update_ret["comment"])
                result["result"] = update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
            else:
                plan_state["tags"] = tags_dict
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "acl": acl,
                    "name": name,
                    "resource_id": resource_id,
                    "create_bucket_configuration": create_bucket_configuration,
                    "grant_full_control": grant_full_control,
                    "grant_read": grant_read,
                    "grant_read_acp": grant_read_acp,
                    "grant_write": grant_write,
                    "grant_write_acp": grant_write_acp,
                    "object_lock_enabled_for_bucket": object_lock_enabled_for_bucket,
                    "object_ownership": object_ownership,
                    "tags": tags_dict,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.s3.bucket", name=name
            )
            return result

        payload = hub.tool.aws.s3.s3_utils.get_create_bucket_payload(
            acl,
            name,
            create_bucket_configuration,
            grant_full_control,
            grant_read,
            grant_read_acp,
            grant_write,
            grant_write_acp,
            object_lock_enabled_for_bucket,
            object_ownership,
        )
        ret = await hub.exec.boto3.client.s3.create_bucket(ctx, **payload)
        if not ret["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.s3.bucket", name=name
        )
        resource_id = name
        if tags_list is not None:
            ret = await hub.exec.boto3.client.s3.put_bucket_tagging(
                ctx,
                Bucket=name,
                Tagging={"TagSet": tags_list},
            )
            if not ret["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = False
                return result

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        convert_ret = await hub.exec.aws.s3.bucket.get(
            ctx, name=name, resource_id=resource_id
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["new_state"] = copy.deepcopy(convert_ret["ret"].get("ret"))
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Delete the specified s3 bucket.

    Args:
        name(str): The Idem name of the s3 bucket.

        resource_id(str, Optional): AWS S3 Bucket name.

    Request Syntax:
        .. code-block:: sls

            idem-name:
              aws.s3.bucket.absent:
                - name: value
                - resource_id: value

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            bucket-5435423646-456464:
              aws.s3.bucket.absent:
                - name: bucket1
                - resource_id: bucket1
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = result[
            "comment"
        ] + hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket", name=name
        )
        return result
    resource = hub.tool.boto3.resource.create(ctx, "s3", "Bucket", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = result[
            "comment"
        ] + hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket", name=name
        )
        return result
    else:
        convert_ret = await hub.exec.aws.s3.bucket.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
            return result
        result["old_state"] = convert_ret["ret"].get("ret")
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.s3.bucket", name=name
            )
            return result
        ret = await hub.exec.boto3.client.s3.delete_bucket(ctx, Bucket=resource_id)
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.s3.bucket", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the AWS bucket in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.s3
    """
    result = {}
    ret = await hub.exec.boto3.client.s3.list_buckets(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe S3 buckets {ret['comment']}")
        return {}

    for bucket in ret["ret"]["Buckets"]:
        resource_id = bucket.get("Name")
        convert_ret = await hub.tool.aws.s3.conversion_utils.convert_raw_s3_to_present(
            ctx, idem_resource_name=resource_id
        )
        if not convert_ret["result"]:
            hub.log.warning(
                f"Could not describe s3 bucket '{resource_id}' with error {convert_ret['comment']}"
            )
            continue
        translated_resource = convert_ret["ret"]
        result[resource_id] = {
            "aws.s3.bucket.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions for AWS API Gateway v2 API resources.
"""


def convert_raw_api_to_present(hub, raw_resource: Dict[str, Any]) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway v2 API resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        raw_resource(Dict[str, Any]): The AWS response to convert.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "ApiEndpoint": "api_endpoint",
            "ApiGatewayManaged": "api_gateway_managed",
            "ApiId": "api_id",
            "ApiKeySelectionExpression": "api_key_selection_expression",
            "CorsConfiguration": "cors_configuration",
            "Description": "description",
            "DisableExecuteApiEndpoint": "disable_execute_api_endpoint",
            "DisableSchemaValidation": "disable_schema_validation",
            "Name": "name",
            "ProtocolType": "protocol_type",
            "RouteSelectionExpression": "route_selection_expression",
            "Tags": "tags",
            "Version": "version",
        }
    )
    resource_translated = {
        "resource_id": raw_resource.get("ApiId"),
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    resource_id: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, None],
) -> Dict[str, Any]:
    r"""
    Updates an AWS API Gateway v2 API resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        resource_id(string): The API resource identifier in Amazon Web Services.
        raw_resource(Dict): Existing resource parameters in Amazon Web Services.
        resource_parameters(Dict): Parameters from SLS file.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)

    parameters = OrderedDict(
        {
            "api_key_selection_expression": "ApiKeySelectionExpression",
            "cors_configuration": "CorsConfiguration",
            "credentials_arn": "CredentialsArn",
            "description": "Description",
            "disable_execute_api_endpoint": "DisableExecuteApiEndpoint",
            "disable_schema_validation": "DisableSchemaValidation",
            "route_key": "RouteKey",
            "route_selection_expression": "RouteSelectionExpression",
            "target": "Target",
            "version": "Version",
        }
    )

    parameters_to_update = {}

    cors_configuration = resource_parameters.get("cors_configuration")
    if cors_configuration is not None:
        update_cors_configuration = False
        old_cors_configuration = raw_resource.get("cors_configuration")

        if cors_configuration.get(
            "AllowCredentials"
        ) is not None and cors_configuration.get(
            "AllowCredentials"
        ) != old_cors_configuration.get(
            "MaxAge"
        ):
            update_cors_configuration = True
        elif (
            not hub.tool.aws.state_comparison_utils.are_lists_identical(
                cors_configuration.get("AllowHeaders"),
                old_cors_configuration.get("AllowHeaders"),
            )
            or not hub.tool.aws.state_comparison_utils.are_lists_identical(
                cors_configuration.get("AllowMethods"),
                old_cors_configuration.get("AllowMethods"),
            )
            or not hub.tool.aws.state_comparison_utils.are_lists_identical(
                cors_configuration.get("AllowOrigins"),
                old_cors_configuration.get("AllowOrigins"),
            )
            or not hub.tool.aws.state_comparison_utils.are_lists_identical(
                cors_configuration.get("ExposeHeaders"),
                old_cors_configuration.get("ExposeHeaders"),
            )
        ):
            update_cors_configuration = True
        elif cors_configuration.get("MaxAge") is not None and cors_configuration.get(
            "MaxAge"
        ) != old_cors_configuration.get("MaxAge"):
            update_cors_configuration = True

        if update_cors_configuration:
            parameters_to_update["CorsConfiguration"] = cors_configuration

        resource_parameters.pop("cors_configuration")

    for key, value in resource_parameters.items():
        if value is not None and value != raw_resource[key]:
            parameters_to_update[parameters[key]] = resource_parameters[key]

    if parameters_to_update:
        result["ret"] = {}
        for parameter_present, parameter_raw in parameters.items():
            if parameter_raw in parameters_to_update:
                result["ret"][parameter_present] = parameters_to_update[parameter_raw]

        if ctx.get("test", False):
            result["comment"] = (
                "Would update parameters: " + ",".join(result["ret"].keys()),
            )
        else:
            update_ret = await hub.exec.boto3.client.apigatewayv2.update_api(
                ctx,
                ApiId=resource_id,
                **parameters_to_update,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result

            result["comment"] = (
                "Updated parameters: " + ",".join(result["ret"].keys()),
            )

    return result



"""State module for managing acl policy for S3 bucket"""
import copy
import json
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    access_control_policy: make_dataclass(
        "AccessControlPolicy",
        [
            (
                "Grants",
                List[
                    make_dataclass(
                        "Grant",
                        [
                            (
                                "Grantee",
                                make_dataclass(
                                    "Grantee",
                                    [
                                        ("DisplayName", str, field(default=None)),
                                        ("EmailAddress", str, field(default=None)),
                                        ("ID", str, field(default=None)),
                                        ("Type", str, field(default=None)),
                                        ("URI", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("Permission", str, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
            (
                "Owner",
                make_dataclass(
                    "Owner",
                    [
                        ("DisplayName", str, field(default=None)),
                        ("ID", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ),
    resource_id: str = None,
) -> Dict[str, Any]:
    """Create/Update the ACL policy for S3 bucket.

    Amazon S3 access control lists (ACLs) enable you to manage access to bucket and objects. Each bucket and object
    has an ACL attached to it as a subresource. It defines which AWS accounts or groups are granted access and the type
    of access. When a request is received against a resource, Amazon S3 checks the corresponding ACL to verify that the
    requester has the necessary access permissions.


    Args:
        name(str):
            The bucket name on which ACL gets apply.

        resource_id(str, optional):
            Name of the S3 bucket.

        access_control_policy(dict):
            Contains the elements that set the ACL permissions for an object per grantee.

            * Grants (List[Dict[str, Any]], optional):
                A list of grants.

                * Grantee (Dict[str, Any], optional):
                    The person being granted permissions.

                    * DisplayName (str, optional):
                        Screen name of the grantee.

                    * EmailAddress (str, optional):
                        Email address of the grantee.  Using email addresses to specify a grantee is only supported in
                        the following Amazon Web Services Regions:    US East (N. Virginia)   US West (N. California)
                        US West (Oregon)    Asia Pacific (Singapore)   Asia Pacific (Sydney)   Asia Pacific (Tokyo)
                        Europe (Ireland)   South America (São Paulo)   For a list of all the Amazon S3 supported Regions
                        and endpoints, see Regions and Endpoints in the Amazon Web Services General Reference.
                    * ID (str, optional):
                        The canonical user ID of the grantee.

                    * Type (str):
                        Type of grantee

                    * URI (str, optional):
                        URI of the grantee group.

                * Permission (str, optional):
                    Specifies the permission given to the grantee.

            * Owner (dict, optional):
                Container for the bucket owner's display name and ID.

                * DisplayName (str, optional):
                    Container for the display name of the owner.

                * ID (str, optional):
                    Container for the ID of the owner.

    Request Syntax:
        .. code-block:: sls

            [bucket_name]:
              aws.s3.bucket_acl.present:
              - name: "string"
              - resource_id: "string"
              - access_control_policy:
                  Grants:
                    - Grantee:
                        DisplayName: [string]
                        EmailAddress: [string]
                        ID: [string]
                        Type: [string]
                        URI: [string]
                      Permission: [string]
                  Owner:
                    DisplayName: [string]
                    ID: [string]
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws_auto.s3.bucket_acl.present:
                - name: value
                - access_control_policy: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False

    if resource_id:
        before_ret = await hub.exec.aws.s3.bucket_acl.get(
            ctx=ctx, resource_id=resource_id, name=name
        )
        if not before_ret["result"] or not before_ret["ret"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

        result["old_state"] = copy.deepcopy(before_ret["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])

        update_bucket_acl = await hub.tool.aws.s3.bucket_acl.update(
            ctx,
            resource_id=resource_id,
            raw_resource=before_ret["ret"],
            access_control_policy=access_control_policy,
        )
        result["comment"] += update_bucket_acl["comment"]
        if not update_bucket_acl["result"]:
            result["result"] = False
            return result

        if not update_bucket_acl["ret"]:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.s3.bucket_acl", name=name
            )

        if update_bucket_acl["ret"] and ctx.get("test", False):
            result["new_state"].update(update_bucket_acl["ret"])
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.s3.bucket_acl", name=name
            )
        resource_updated = resource_updated or bool(update_bucket_acl["ret"])
        if resource_updated and ctx.get("test", False):
            return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "access_control_policy": access_control_policy,
                    "resource_id": name,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.s3.bucket_acl", name=name
            )
            return result

        create_ret = await hub.exec.boto3.client.s3.put_bucket_acl(
            ctx,
            Bucket=name,
            AccessControlPolicy=access_control_policy,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.s3.bucket_acl", name=name
        )
        resource_id = name

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.aws.s3.bucket_acl.get(
            ctx=ctx, resource_id=resource_id, name=name
        )
        if not after_ret["result"] or not after_ret["ret"]:
            result["result"] = False
            result["comment"] = result["comment"] + tuple(after_ret["comment"])
            return result
        result["new_state"] = after_ret["ret"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes associated ACL policy for an S3 bucket.

    This action cannot be possible as AWS does not provide an API to dissociate the ACL from S3 bucket,
    The ACL policy will be deleted post the bucket gets deleted.

    Args:
        name(str): An S3 bucket name.
        resource_id(str, Optional): An identifier of the resource in the provider.

    Request Syntax:
        .. code-block:: sls

            [bucket_name]:
              aws.s3.bucket_acl.absent:
                - name: "string"
                - resource_id: "string"
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.s3.bucket_acl.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = await hub.exec.aws.s3.bucket_acl.get(
        ctx=ctx, resource_id=resource_id, name=name
    )

    if not before_ret["result"] or not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket_acl", name=name
        )
    else:
        result["old_state"] = copy.deepcopy(before_ret["ret"])
        result[
            "comment"
        ] = f"This action cannot be possible as AWS does not provide an API to dissociate the ACL from S3 bucket, The ACL policy will be deleted post '{name}' bucket gets deleted."

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Gets information about the ACL policy attached to S3 bucket.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.s3.bucket_acl
    """

    result = {}
    # To describe the ACL policy of all the buckets, we first need to list all the buckets,
    # then get the ACL policy for each bucket
    ret = await hub.exec.boto3.client.s3.list_buckets(ctx)
    if not ret["result"]:
        hub.log.error(f"Could not describe bucket_acl {ret['comment']}")
        return {}

    for bucket in ret["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")
        # get acl policy for each bucket
        acl_policy = await hub.exec.aws.s3.bucket_acl.get(
            ctx=ctx, resource_id=bucket_name, name=bucket_name
        )
        if acl_policy["result"]:
            policy_translated = json.loads(json.dumps(acl_policy["ret"]))

            result[bucket_name + "-bucket_acl"] = {
                "aws.s3.bucket_acl.present": [policy_translated]
            }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.rds.copy_db_parameter_group
hub.exec.boto3.client.rds.create_db_parameter_group
hub.exec.boto3.client.rds.delete_db_parameter_group
hub.exec.boto3.client.rds.describe_db_parameter_groups
hub.exec.boto3.client.rds.modify_db_parameter_group
hub.exec.boto3.client.rds.reset_db_parameter_group
"""
import copy
from collections import OrderedDict
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_parameter_group_family: str,
    description: str = None,
    resource_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
    parameters: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new DB parameter group. A DB parameter group is initially created with the default parameters for the
    database engine used by the DB instance. To provide custom values for any of the parameters, you can specify parameters.
    Once you've created a DB parameter group, you can associate it with your DB instance. When you associate a new DB parameter group with a
    running DB instance, you need to reboot the DB instance without failover for the new DB parameter group and
    associated settings to take effect. After you create a DB parameter
    group, you should wait at least 5 minutes before creating your first DB instance that uses that DB parameter
    group as the default parameter group. This allows Amazon RDS to fully complete the create action before the
    parameter group is used as the default for a new DB instance. This is especially important for parameters that
    are critical when creating the default database for a DB instance, such as the character set for the default
    database defined by the character_set_database parameter. You can use the Parameter Groups option of the Amazon
    RDS console to verify that your DB parameter group has been created or modified.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        db_parameter_group_family(Text): The DB parameter group family name. A DB parameter group can be associated with one and only one
            DB parameter group family, and can be applied only to a DB instance running a database engine
            and engine version compatible with that DB parameter group family. To list all of the available
            parameter group families for a DB engine, use the following command:  aws rds describe-db-
            engine-versions --query "DBEngineVersions[].DBParameterGroupFamily" --engine <engine>  For
            example, to list all of the available parameter group families for the MySQL DB engine, use the
            following command:  aws rds describe-db-engine-versions --query
            "DBEngineVersions[].DBParameterGroupFamily" --engine mysql   The output contains duplicates.
            The following are the valid DB engine values:    aurora (for MySQL 5.6-compatible Aurora)
            aurora-mysql (for MySQL 5.7-compatible and MySQL 8.0-compatible Aurora)    aurora-postgresql
            mariadb     mysql     oracle-ee     oracle-ee-cdb     oracle-se2     oracle-se2-cdb     postgres
            sqlserver-ee     sqlserver-se     sqlserver-ex     sqlserver-web.
        description(Text): The description for the DB parameter group.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the DB parameter group.
            Each tag consists of a key name and an associated value. Defaults to None.
            * Key (str, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * Value(str, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        parameters(List, optional): parameters in the DB parameter group. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.rds.db_parameter_group.present:
              - name: 'test'
              - resource_id: 'test'
              - db_parameter_group_family: 'aurora-mysql5.7'
              - description: 'test'
              - tags:
                 - Key: 'test'
                   Value: 'test1'
                 - Key: 'test2'
                   Value: 'test2'
              - parameters:
                 - name: 'aurora_disable_hash_join'
                   value: '1'
                   apply_method: 'immediate'
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.rds.describe_db_parameter_groups(
            ctx,
            DBParameterGroupName=resource_id,
        )
    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )
    if before:
        ret_tags = None
        tags_list = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx,
            ResourceName=before["ret"]["DBParameterGroups"][0].get(
                "DBParameterGroupArn"
            ),
        )
        if tags_list["result"]:
            ret_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                tags_list["ret"]["TagList"]
            )

        parameters_list = await hub.exec.boto3.client.rds.describe_db_parameters(
            ctx, DBParameterGroupName=resource_id
        )
        ret_parameters = None
        if parameters_list["result"]:
            ret_parameters = parameters_list["ret"]["Parameters"]
        result[
            "old_state"
        ] = await hub.tool.aws.rds.conversion_utils.convert_db_parameter_group_to_present(
            raw_resource=before["ret"]["DBParameterGroups"][0],
            idem_resource_name=name,
            parameters=ret_parameters,
            tags=ret_tags,
        )
        plan_state = copy.deepcopy(result["old_state"])

        if tags is not None:
            # Update tags
            update_ret = await hub.exec.aws.rds.tag.update_rds_tags(
                ctx=ctx,
                resource_arn=result["old_state"].get("db_parameter_group_arn"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]
            resource_updated = resource_updated or update_ret["result"]
            if ctx.get("test", False) and update_ret["ret"] is not None:
                plan_state["tags"] = update_ret["ret"]
        if parameters is not None:
            update_ret = await hub.tool.aws.rds.parameters.modify_db_group_parameters(
                ctx=ctx,
                resource_name=result["old_state"].get("resource_id"),
                old_parameters=result["old_state"].get("parameters"),
                new_parameters=parameters,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if ctx.get("test", False) and update_ret["ret"] is not None:
                plan_state["parameters"] = update_ret["ret"].get("parameters")
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "db_parameter_group_family": db_parameter_group_family,
                    "description": description,
                    "tags": tags,
                    "parameters": parameters,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.rds.db_parameter_group", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.rds.create_db_parameter_group(
                ctx,
                **{
                    "DBParameterGroupName": name,
                    "DBParameterGroupFamily": db_parameter_group_family,
                    "Description": description,
                    "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                },
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

            resource_id = ret["ret"]["DBParameterGroup"]["DBParameterGroupName"]
            result["new_state"] = {"name": name, "resource_id": resource_id}
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.rds.db_parameter_group", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    # Modify the db group parameters
    if parameters is not None:
        ret_parameter = await hub.exec.boto3.client.rds.describe_db_parameters(
            ctx, DBParameterGroupName=resource_id
        )
        db_parameters = OrderedDict(
            {"ParameterName": "name", "ParameterValue": "value"}
        )
        list_parameters = []
        for parameter in ret_parameter["ret"]["Parameters"]:
            update_parameters = {}
            for parameter_old_key, parameter_new_key in db_parameters.items():
                if parameter.get(parameter_old_key) is not None:
                    update_parameters[parameter_new_key] = parameter.get(
                        parameter_old_key
                    )
            list_parameters.append(update_parameters)
        update_ret = await hub.tool.aws.rds.parameters.modify_db_group_parameters(
            ctx=ctx,
            resource_name=resource_id,
            old_parameters=list_parameters,
            new_parameters=parameters,
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = result["result"] and update_ret["result"]

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.rds.describe_db_parameter_groups(
                ctx,
                DBParameterGroupName=resource_id,
            )
            ret_tags = None
            tags_list = await hub.exec.boto3.client.rds.list_tags_for_resource(
                ctx,
                ResourceName=after["ret"]["DBParameterGroups"][0].get(
                    "DBParameterGroupArn"
                ),
            )
            if tags_list["result"]:
                ret_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    tags_list["ret"]["TagList"]
                )

            parameters_list = await hub.exec.boto3.client.rds.describe_db_parameters(
                ctx, DBParameterGroupName=resource_id
            )
            ret_parameters = None
            if parameters_list["result"]:
                ret_parameters = parameters_list["ret"]["Parameters"]

            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result[
                "new_state"
            ] = await hub.tool.aws.rds.conversion_utils.convert_db_parameter_group_to_present(
                raw_resource=after["ret"]["DBParameterGroups"][0],
                idem_resource_name=name,
                parameters=ret_parameters,
                tags=ret_tags,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.rds.db_parameter_group", name=name
            )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a specified DB parameter group. The DB parameter group to be deleted can't be associated with any DB
    instances.

    Args:
        resource_id(Text): An identifier of the resource in the provider.
        name(Text): The name of the DB parameter group. Constraints:   Must be the name of an existing DB parameter
            group   You can't delete a default DB parameter group   Can't be associated with any DB
            instances.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.rds.db_parameter_group.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_parameter_group", name=name
        )
        return result
    before = await hub.exec.boto3.client.rds.describe_db_parameter_groups(
        ctx,
        DBParameterGroupName=resource_id,
    )
    if not before or not before["ret"].get("DBParameterGroups"):
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_parameter_group", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.rds.conversion_utils.convert_db_parameter_group_to_present(
            raw_resource=before["ret"]["DBParameterGroups"][0], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.rds.db_parameter_group", name=name
        )
        return result
    else:
        tags = None
        tags_list = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx,
            ResourceName=before["ret"]["DBParameterGroups"][0].get(
                "DBParameterGroupArn"
            ),
        )
        if tags_list["result"]:
            tags = tags_list["ret"]["TagList"]

        ret_parameter = await hub.exec.boto3.client.rds.describe_db_parameters(
            ctx, DBParameterGroupName=resource_id
        )
        parameters = None
        if ret_parameter["result"]:
            parameters = ret_parameter["ret"]["Parameters"]
        result[
            "old_state"
        ] = await hub.tool.aws.rds.conversion_utils.convert_db_parameter_group_to_present(
            raw_resource=before["ret"]["DBParameterGroups"][0],
            idem_resource_name=name,
            parameters=parameters,
            tags=tags,
        )
        try:
            ret = await hub.exec.boto3.client.rds.delete_db_parameter_group(
                ctx,
                DBParameterGroupName=name,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.rds.db_parameter_group", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of DBParameterGroup descriptions.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.rds.db_parameter_group
    """

    result = {}
    ret = await hub.exec.boto3.client.rds.describe_db_parameter_groups(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe db_parameter_group {ret['comment']}")
        return {}

    for db_parameter_group in ret["ret"]["DBParameterGroups"]:
        resource_id = db_parameter_group.get("DBParameterGroupName")
        tags = None
        tags_list = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx, ResourceName=db_parameter_group.get("DBParameterGroupArn")
        )
        if tags_list["result"]:
            tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                tags_list["ret"]["TagList"]
            )

        ret_parameter = await hub.exec.boto3.client.rds.describe_db_parameters(
            ctx, DBParameterGroupName=resource_id
        )
        parameters = None
        if ret_parameter["result"]:
            parameters = ret_parameter["ret"]["Parameters"]
        resource_translated = await hub.tool.aws.rds.conversion_utils.convert_db_parameter_group_to_present(
            raw_resource=db_parameter_group,
            idem_resource_name=resource_id,
            tags=tags,
            parameters=parameters,
        )
        result[resource_id] = {
            "aws.rds.db_parameter_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""State module for managing AWS Budget Actions."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    budget_name: str,
    notification_type: str,
    action_type: str,
    action_threshold: make_dataclass(
        "ActionThreshold",
        [("ActionThresholdValue", float), ("ActionThresholdType", str)],
    ),
    definition: make_dataclass(
        "Definition",
        [
            (
                "IamActionDefinition",
                make_dataclass(
                    "IamActionDefinition",
                    [
                        ("PolicyArn", str),
                        ("Roles", List[str], field(default=None)),
                        ("Groups", List[str], field(default=None)),
                        ("Users", List[str], field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "ScpActionDefinition",
                make_dataclass(
                    "ScpActionDefinition", [("PolicyId", str), ("TargetIds", List[str])]
                ),
                field(default=None),
            ),
            (
                "SsmActionDefinition",
                make_dataclass(
                    "SsmActionDefinition",
                    [
                        ("ActionSubType", str),
                        ("Region", str),
                        ("InstanceIds", List[str]),
                    ],
                ),
                field(default=None),
            ),
        ],
    ),
    execution_role_arn: str,
    approval_model: str = None,
    subscribers: List[
        make_dataclass("Subscriber", [("SubscriptionType", str), ("Address", str)])
    ] = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Creates an AWS Budget Action with the requested name and rules.

    Args:
        name(str):
            The unique name of the Budget Action.
        budget_name(str):
            A string that represents the budget name. The ":" and "" characters aren't allowed.
        notification_type(str):
            The type of a notification. It must be ACTUAL or FORECASTED.
        action_type(str):
            The type of action. This defines the type of tasks that can be carried out by this action. This field also
            determines the format for definition.
        action_threshold(dict[str, Any]):
            The trigger threshold of the action.

            * ActionThresholdValue (float):
                The threshold of a notification.

            * ActionThresholdType (str):
                The type of threshold for a notification.
        definition(dict[str, Any]):
            Specifies all of the type-specific parameters.

            * IamActionDefinition (dict[str, Any], Optional):
                The Identity and Access Management (IAM) action definition details.

                * PolicyArn (str):
                    The Amazon Resource Name (ARN) of the policy to be attached.

                * Roles (list[str], Optional):
                    A list of roles to be attached. There must be at least one role.

                * Groups (list[str], Optional):
                    A list of groups to be attached. There must be at least one group.

                * Users (list[str], Optional):
                    A list of users to be attached. There must be at least one user.

            * ScpActionDefinition (dict[str, Any], Optional):
                The service control policies (SCPs) action definition details.

                * PolicyId (str):
                    The policy ID attached.

                * TargetIds (list[str]):
                    A list of target IDs.

            * SsmActionDefinition (dict[str, Any], Optional):
                The Amazon Web Services Systems Manager (SSM) action definition details.

                * ActionSubType (str):
                    The action subType.

                * Region (str):
                    The Region to run the SSM document.

                * InstanceIds (list[str]):
                    The EC2 and RDS instance IDs.
        execution_role_arn(str):
            The role passed for action execution and reversion. Roles and actions must be in the same account.
        approval_model (str):
            This specifies if the action needs manual or automatic approval.
        subscribers(list[dict[str, Any]]):
            A list of subscribers.

            The subscriber to a budget notification consists of a subscription type and either an Amazon SNS topic or
            an email address.

            For example, an email subscriber has the following parameters:
              * A `subscriptionType` of `EMAIL`
              * An `address` of `example@example.com`

            * SubscriptionType (str):
                The type of notification that Amazon Web Services sends to a subscriber.

            * Address (str):
                The address that Amazon Web Services sends budget notifications to, either an SNS topic or an
                email.

                When you create a subscriber, the value of `Address` can't contain line breaks.
        resource_id(str, Optional):
            Action ID to identify the resource.

    Request Syntax:
      .. code-block:: sls

        [action-resource-id]:
          aws.budgets.budget_action.present:
            - resource_id: "string"
            - budget_name: "string"
            - notification_type: "string"
            - action_type: "string"
            - action_threshold:
                ActionThresholdType: "string"
                ActionThresholdValue: int
            - definition:
                IamActionDefinition:
                  PolicyArn: "string"
                  Users:
                    - "string"
            - execution_role_arn: "string"
            - approval_model: "string"
            - status: "string"
            - subscribers:
                - Address: "string"
                  SubscriptionType: "string"


    Returns:
        Dict[str, str]

    Examples:
        .. code-block:: sls

            12345a12-126d-1234-1ab2-1f4dca2a1234:
              aws.budgets.budget_action.present:
                - name: 12345a12-126d-1234-1ab2-1f4dca2a1234
                - resource_id: 12345a12-126d-1234-1ab2-1f4dca2a1234
                - budget_name: test_budget
                - notification_type: ACTUAL
                - action_type: APPLY_IAM_POLICY
                - action_threshold:
                    ActionThresholdType: PERCENTAGE
                    ActionThresholdValue: 123.0
                - definition:
                    IamActionDefinition:
                      PolicyArn: arn:aws:iam::123456789101:policy/budget_policy
                      Users:
                        - plokare
                - execution_role_arn: arn:aws:iam::123456789101:role/budget_role
                - approval_model: AUTOMATIC
                - status: STANDBY
                - subscribers:
                    - Address: abc@test.com
                      SubscriptionType: EMAIL

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = None
    resource_updated = False
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]

    if resource_id:
        before = await hub.exec.boto3.client.budgets.describe_budget_action(
            ctx, AccountId=account_id, BudgetName=budget_name, ActionId=resource_id
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

    if before:
        result[
            "old_state"
        ] = hub.tool.aws.budgets.conversion_utils.convert_raw_budget_action_to_present(
            ctx, raw_resource=before["ret"]["Action"], idem_resource_name=name
        )
        plan_state = copy.deepcopy(result["old_state"])

        update_ret = await hub.tool.aws.budgets.budget_action.update(
            ctx,
            before=before["ret"]["Action"],
            account_id=account_id,
            notification_type=notification_type,
            action_threshold=action_threshold,
            definition=definition,
            execution_role_arn=execution_role_arn,
            approval_model=approval_model,
            subscribers=subscribers,
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            plan_state = update_ret["ret"]

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.budgets.budget_action", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.budgets.budget_action", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "budget_name": budget_name,
                    "notification_type": notification_type,
                    "action_type": action_type,
                    "action_threshold": action_threshold,
                    "definition": definition,
                    "execution_role_arn": execution_role_arn,
                    "approval_model": approval_model,
                    "subscribers": subscribers,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.budgets.budget_action", name=name
            )
            return result

        ret = await hub.exec.boto3.client.budgets.create_budget_action(
            ctx,
            AccountId=account_id,
            BudgetName=budget_name,
            NotificationType=notification_type,
            ActionType=action_type,
            ActionThreshold=action_threshold,
            Definition=definition,
            ExecutionRoleArn=execution_role_arn,
            ApprovalModel=approval_model,
            Subscribers=subscribers,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.budgets.budget_action", name=name
        )
        resource_id = ret["ret"]["ActionId"]

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.budgets.describe_budget_action(
                ctx, AccountId=account_id, BudgetName=budget_name, ActionId=resource_id
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result[
                "new_state"
            ] = hub.tool.aws.budgets.conversion_utils.convert_raw_budget_action_to_present(
                ctx,
                raw_resource=after["ret"]["Action"],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub, ctx, name: str, budget_name: str, resource_id: str = None
) -> Dict[str, Any]:
    """Deletes an AWS Budget Action.

    Args:
        name(str): An Idem name of the AWS Budget Action.
        budget_name(str): Budget Name
        resource_id(str, Optional): Budget Action ID to identify the resource.
            Idem automatically considers this resource being absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            [action-resource-id]:
              aws.budgets.budget_action.absent:
                - name: value
                - budget_name: value
                - resource_id: value

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.budgets.budget_action.absent:
                - name: value
                - budget_name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.budgets.budget_action", name=name
        )
        return result

    before = await hub.exec.boto3.client.budgets.describe_budget_action(
        ctx, AccountId=account_id, BudgetName=budget_name, ActionId=resource_id
    )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.budgets.budget_action", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.budgets.conversion_utils.convert_raw_budget_action_to_present(
            ctx, raw_resource=before["ret"]["Action"], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.budgets.budget_action", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.budgets.conversion_utils.convert_raw_budget_action_to_present(
            ctx, raw_resource=before["ret"]["Action"], idem_resource_name=name
        )
        ret = await hub.exec.boto3.client.budgets.delete_budget_action(
            ctx, AccountId=account_id, BudgetName=budget_name, ActionId=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.budgets.budget_action", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Lists the AWS Budget Actions.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.budgets.budget_action
    """
    result = {}

    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]

    ret = await hub.exec.boto3.client.budgets.describe_budget_actions_for_account(
        ctx, AccountId=account_id
    )

    if not ret["result"]:
        hub.log.debug(f"Could not list Budget Actions {ret['comment']}")
        return {}

    for action in ret["ret"]["Actions"]:
        resource_id = action.get("ActionId")

        resource_translated = (
            hub.tool.aws.budgets.conversion_utils.convert_raw_budget_action_to_present(
                ctx,
                raw_resource=action,
                idem_resource_name=resource_id,
            )
        )
        result[resource_id] = {
            "aws.budgets.budget_action.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS CloudwatchLog Resource Policy to present input format.
"""


def convert_raw_log_group_resource_policy_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("policyName")
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "policy_document": raw_resource.get("policyDocument"),
    }
    return resource_translated



"""State module for managing EC2 Volumes"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    availability_zone: str,
    resource_id: str = None,
    encrypted: bool = None,
    iops: int = None,
    kms_key_id: str = None,
    outpost_arn: str = None,
    size: int = None,
    snapshot_id: str = None,
    volume_type: str = None,
    multi_attach_enabled: bool = None,
    throughput: int = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """Register an AWS Volume

    Creates an EBS volume that can be attached to an instance in the same Availability Zone. You can create a new
    empty volume or restore a volume from an EBS snapshot. Any Amazon Web Services Marketplace product codes from
    the snapshot are propagated to the volume. You can create encrypted volumes. Encrypted volumes must be attached
    to instances that support Amazon EBS encryption. Volumes that are created from encrypted snapshots are also
    automatically encrypted. For more information, see Amazon EBS encryption in the Amazon Elastic Compute Cloud
    User Guide. You can tag your volumes during creation. For more information, see Tag your Amazon EC2 resources in
    the Amazon Elastic Compute Cloud User Guide. For more information, see Create an Amazon EBS volume in the Amazon
    Elastic Compute Cloud User Guide.

    Args:
        name (str):
            An Idem name of the volume resource.

        resource_id (str, optional):
            The AWS volume_id of the Volume. Defaults to None.

        availability_zone (str):
            The Availability Zone in which to create the volume.

        encrypted (bool, optional):
            Indicates whether the volume should be encrypted. The effect of setting the encryption state to
            true depends on the volume origin (new or from a snapshot), starting encryption state,
            ownership, and whether encryption by default is enabled. For more information, see Encryption by
            default in the Amazon Elastic Compute Cloud User Guide. Encrypted Amazon EBS volumes must be
            attached to instances that support Amazon EBS encryption. For more information, see Supported
            instance types. Defaults to None.

        iops (int, optional):
            The number of I/O operations per second (IOPS). For gp3, io1, and io2 volumes, this represents
            the number of IOPS that are provisioned for the volume. For gp2 volumes, this represents the
            baseline performance of the volume and the rate at which the volume accumulates I/O credits for
            bursting. The following are the supported values for each volume type:    gp3: 3,000-16,000 IOPS
            io1: 100-64,000 IOPS    io2: 100-64,000 IOPS    io1 and io2 volumes support up to 64,000 IOPS
            only on Instances built on the Nitro System. Other instance families support performance up to
            32,000 IOPS. This parameter is required for io1 and io2 volumes. The default for gp3 volumes is
            3,000 IOPS. This parameter is not supported for gp2, st1, sc1, or standard volumes. Defaults to None.

        kms_key_id (str, optional):
            The identifier of the Key Management Service (KMS) KMS key to use for Amazon EBS encryption. If
            this parameter is not specified, your KMS key for Amazon EBS is used. If KmsKeyId is specified,
            the encrypted state must be true. You can specify the KMS key using any of the following:   Key
            ID. For example, 1234abcd-12ab-34cd-56ef-1234567890ab.   Key alias. For example,
            alias/ExampleAlias.   Key ARN. For example, arn:aws:kms:us-
            east-1:012345678910:key/1234abcd-12ab-34cd-56ef-1234567890ab.   Alias ARN. For example,
            arn:aws:kms:us-east-1:012345678910:alias/ExampleAlias.   Amazon Web Services authenticates the
            KMS key asynchronously. Therefore, if you specify an ID, alias, or ARN that is not valid, the
            action can appear to complete, but eventually fails. Defaults to None.

        outpost_arn (str, optional):
            The Amazon Resource Name (ARN) of the Outpost. Defaults to None.

        size (int, optional):
            The size of the volume, in GiBs. You must specify either a snapshot ID or a volume size. If you
            specify a snapshot, the default is the snapshot size. You can specify a volume size that is
            equal to or larger than the snapshot size. The following are the supported volumes sizes for
            each volume type:    gp2 and gp3: 1-16,384    io1 and io2: 4-16,384    st1 and sc1: 125-16,384
            standard: 1-1,024. Defaults to None.

        snapshot_id (str, optional):
            The snapshot from which to create the volume. You must specify either a snapshot ID or a volume
            size. Defaults to None.

        volume_type (str, optional): ]
            The volume type. This parameter can be one of the following values:   General Purpose SSD: gp2 |
            gp3    Provisioned IOPS SSD: io1 | io2    Throughput Optimized HDD: st1    Cold HDD: sc1
            Magnetic: standard    For more information, see Amazon EBS volume types in the Amazon Elastic
            Compute Cloud User Guide. Default: gp2. Defaults to None.

        multi_attach_enabled (bool, optional):
            Indicates whether to enable Amazon EBS Multi-Attach. If you enable Multi-Attach, you can attach
            the volume to up to 16 Instances built on the Nitro System in the same Availability Zone. This
            parameter is supported with io1 and io2 volumes only. For more information, see  Amazon EBS
            Multi-Attach in the Amazon Elastic Compute Cloud User Guide. Defaults to None.

        throughput (int, optional):
            The throughput to provision for a volume, with a maximum of 1,000 MiB/s. This parameter is valid
            only for gp3 volumes. Valid Range: Minimum value of 125. Maximum value of 1000. Defaults to None.

        tags (dict, optional):
            dict in the format of ``{tag-key: tag-value}``. Defaults to None.

            * Key (str):
                The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .

            * Value (str):
                The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws_auto.ec2.volume.present:
                - name: some-volume-ARN
                - snapshot_id: some-snapshot-ARN
                - volume_type: standard
                - availability_zone: us-east-1a
                - size: 1
                - encrypted: True
                - tags:
                  - Key: name
                    Value: some-volume-ARN
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.ec2.volume.get(ctx=ctx, resource_id=resource_id)
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=before["ret"]["resource_id"],
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]
            resource_updated = resource_updated or update_ret["result"]
            if ctx.get("test", False) and update_ret["result"]:
                plan_state["tags"] = update_ret["ret"]
        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.ec2.volume", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.ec2.volume", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "availability_zone": availability_zone,
                    "encrypted": encrypted,
                    "iops": iops,
                    "kms_key_id": kms_key_id,
                    "outpost_arn": outpost_arn,
                    "size": size,
                    "snapshot_id": snapshot_id,
                    "volume_type": volume_type,
                    "tags": tags,
                    "multi_attach_enabled": multi_attach_enabled,
                    "throughput": throughput,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.volume", name=name
            )
            return result

        ret = await hub.exec.boto3.client.ec2.create_volume(
            ctx,
            AvailabilityZone=availability_zone,
            Encrypted=encrypted,
            Iops=iops,
            KmsKeyId=kms_key_id,
            OutpostArn=outpost_arn,
            Size=size,
            SnapshotId=snapshot_id,
            VolumeType=volume_type,
            Throughput=throughput,
            MultiAttachEnabled=multi_attach_enabled,
            TagSpecifications=[
                {
                    "ResourceType": "volume",
                    "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                }
            ]
            if tags
            else None,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["VolumeId"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.volume", name=name
        )

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.aws.ec2.volume.get(
                ctx=ctx, name=name, resource_id=resource_id
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    """Deregisters the specified Volume

    Deletes the specified EBS volume. The volume must be in the available state (not attached to an instance). The
    volume can remain in the deleting state for several minutes. For more information, see Delete an Amazon EBS
    volume in the Amazon Elastic Compute Cloud User Guide.

    Args:
        name (str): An Idem name of the resource.
        resource_id (str): Volume ID.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws_auto.ec2.volume.absent:
                - name: idem-test-volume
                - resource_id: volume-123456789
                - volume_id: volume-123456789
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.volume", name=name
        )
        return result

    before = await hub.exec.aws.ec2.volume.get(
        ctx=ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.volume", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.volume", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.ec2.delete_volume(ctx, VolumeId=resource_id)
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ec2.volume", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Describes the specified EBS volumes or all of your EBS volumes. If you are describing a long list of volumes, we
    recommend that you paginate the output to make the list more manageable. The MaxResults parameter sets the
    maximum number of results returned in a single page. If the list of results exceeds your MaxResults value, then
    that number of results is returned along with a NextToken value that can be passed to a subsequent
    DescribeVolumes request to retrieve the remaining results. For more information about EBS volumes, see Amazon
    EBS volumes in the Amazon Elastic Compute Cloud User Guide.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws_auto.ec2.volume

    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_volumes(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe volume {ret['comment']}")
        return {}

    for volume in ret["ret"]["Volumes"]:
        # Including fields to match the 'present' function parameters
        resource_id = volume.get("VolumeId")
        resource_translated = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_volume_to_present(
                raw_resource=volume, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.volume.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""State module for managing Amazon Elasticache Subnet Groups."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]
TREQ = {
    "present": {
        "require": [
            "aws.ec2.subnet.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    subnet_ids: List[str],
    cache_subnet_group_description: str,
    resource_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    """Creates an AWS Elasticache Subnet Group.

    Args:
        name(str):
            An Idem name of the resource
        resource_id(str, Optional):
            The name of the AWS Elasticache Subnet Group in Amazon Web Services.
        cache_subnet_group_description(str):
            A description for the cache subnet group.
        subnet_ids(list):
            A list of VPC subnet IDs for the cache subnet group.
        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"Key": tag-key, "Value": tag-value}]`` to associate with the subnet group.

            * Key (*str*):
                The key of the tag.
            * Value (*str*):
                The value of the tag.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_elasticache_cache_subnet_group]:
          aws.elasticache.cache_subnet_group.present:
            - name: 'string'
            - resource_id: 'string'
            - cache_subnet_group_description: 'string'
            - subnet_id:
                - 'string'
            - tags:
                - Key: 'string'
                  Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_elasticache_cache_subnet_group:
              aws.elasticache.cache_subnet_group.present:
                - name: 'idem_test_cache_subnet_group'
                - cache_subnet_group_description: 'My Elasticache Subnet Group'
                - subnet_id:
                  - 'subnet-12345678'
                - tags:
                    - Key: 'provider'
                      Value: 'idem'
    """
    result = dict(comment=(), name=name, old_state=None, new_state=None, result=True)
    before = None
    resource_updated = False
    are_tags_updated = False
    resource_arn = str
    if resource_id:
        before = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
            ctx, CacheSubnetGroupName=resource_id
        )
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    try:
        if before and before["result"] and before["ret"]["CacheSubnetGroups"]:
            convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
                ctx=ctx,
                raw_resource=before["ret"]["CacheSubnetGroups"][0],
                idem_resource_name=name,
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["old_state"] = convert_ret["ret"]
            plan_state = copy.deepcopy(result["old_state"])
            if cache_subnet_group_description != result["old_state"].get(
                "cache_subnet_group_description"
            ) or not hub.tool.aws.state_comparison_utils.are_lists_identical(
                subnet_ids, result["old_state"].get("subnet_ids")
            ):
                if not ctx.get("test", False):
                    update_ret = await hub.exec.boto3.client.elasticache.modify_cache_subnet_group(
                        ctx=ctx,
                        CacheSubnetGroupName=name,
                        CacheSubnetGroupDescription=cache_subnet_group_description,
                        SubnetIds=subnet_ids,
                    )
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = update_ret["result"]
                    resource_updated = resource_updated or bool(update_ret["ret"])
                else:
                    if cache_subnet_group_description is not None:
                        plan_state[
                            "cache_subnet_group_description"
                        ] = cache_subnet_group_description
                    if subnet_ids is not None:
                        plan_state["subnet_ids"] = subnet_ids
            resource_arn = before["ret"]["CacheSubnetGroups"][0]["ARN"]
            if tags is not None and tags != result["old_state"].get("tags"):
                # Update tags
                update_ret = (
                    await hub.tool.aws.elasticache.elasticache_utils.update_tags(
                        ctx=ctx,
                        resource_arn=resource_arn,
                        old_tags=result["old_state"].get("tags"),
                        new_tags=tags,
                    )
                )
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = result["result"] and update_ret["result"]
                are_tags_updated = bool(update_ret["result"])
                resource_updated = resource_updated or are_tags_updated
                if ctx.get("test", False) and are_tags_updated:
                    plan_state["tags"] = update_ret["ret"]
                    result["comment"] = result["comment"] + (
                        f"Would update tags for aws.elasticache.cache_subnet_group {name}",
                    )
        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "cache_subnet_group_description:": cache_subnet_group_description,
                        "subnet_ids": subnet_ids,
                        "tags": tags,
                    },
                )
                result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.elasticache.cache_subnet_group", name=name
                )
                return result

            ret = await hub.exec.boto3.client.elasticache.create_cache_subnet_group(
                ctx,
                CacheSubnetGroupName=name,
                CacheSubnetGroupDescription=cache_subnet_group_description,
                SubnetIds=subnet_ids,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            resource_id = name
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_arn = ret["ret"]["CacheSubnetGroup"]["ARN"]
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.elasticache.cache_subnet_group", name=name
            )
    except hub.tool.boto3.exception.ClientError as e:
        result["result"] = False
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = (
                await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
                    ctx, CacheSubnetGroupName=resource_id
                )
            )
            if after and after.get("ret"):
                convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
                    ctx=ctx,
                    raw_resource=after["ret"]["CacheSubnetGroups"][0],
                    idem_resource_name=name,
                )
                result["result"] = convert_ret["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + convert_ret["comment"]
                result["new_state"] = convert_ret.get("ret")
            else:
                result["result"] = result["result"] and after["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + after["comment"]
                    return result
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes the specified AWS Elasticache Subnet Group.

    .. warning::
        You cannot delete a default cache subnet group or one that is associated with any clusters.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            The name of the AWS Elasticache Subnet Group in Amazon Web Services.

            .. warning::
              Idem automatically considers this resource being absent if this field is not specified.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_elasticache_cache_subnet_group]:
          aws.elasticache.cache_subnet_group.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_elasticache_cache_subnet_group:
              aws.elasticache.cache_subnet_group.absent:
                - name: 'idem_test_cache_subnet_group'
                - resource_id: 'my-subnet-group'
    """
    result = dict(comment=(), name=name, old_state=None, new_state=None, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.elasticache.cache_subnet_group", name=name
        )
        return result

    before = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
        ctx, CacheSubnetGroupName=resource_id
    )
    if not before or not before["result"] or not before["ret"].get("CacheSubnetGroups"):
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.elasticache.cache_subnet_group", name=name
        )
        return result
    else:
        convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
            ctx=ctx,
            raw_resource=before["ret"]["CacheSubnetGroups"][0],
            idem_resource_name=name,
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.elasticache.cache_subnet_group", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.elasticache.delete_cache_subnet_group(
                ctx, CacheSubnetGroupName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.elasticache.cache_subnet_group", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes Elasticache Subnet Groups in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.elasticache.cache_subnet_group
    """
    result = {}
    ret = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe cache_subnet_groups {ret['comment']}")
        return {}

    for resource in ret["ret"]["CacheSubnetGroups"]:
        resource_id = resource.get("CacheSubnetGroupName")
        convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
            ctx=ctx,
            raw_resource=resource,
            idem_resource_name=resource_id,
        )
        if not convert_ret["result"]:
            hub.log.warning(
                f"Could not describe elasticache subnet group '{resource_id}' with error {convert_ret['comment']}"
            )
            continue
        resource_translated = convert_ret.get("ret")
        result[resource_id] = {
            "aws.elasticache.cache_subnet_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result
